from typing import List, Any, Union
import uuid
import weaviate
import numpy as np

from .online_store import VectorStoreTable, VectorStore, ValueType, VectorType
from ..resources import WeaviateConfig
from ..enums import ScalarType


class WeaviateOnlineTable(VectorStoreTable):
    def __init__(self, client, class_name: str, vector_type: ValueType):
        self.client = client
        self.class_name = class_name
        self.vector_type = vector_type

    def set(self, entity: str, value: Union[List[float], np.ndarray]):
        id = self._create_id(entity)
        properties = {"entity": entity}
        if self.client.data_object.exists(uuid=id, class_name=self.class_name):
            self.client.data_object.replace(
                data_object=properties,
                class_name=self.class_name,
                uuid=id,
                vector=value,
            )
        else:
            self.client.data_object.create(
                data_object=properties,
                class_name=self.class_name,
                uuid=id,
                vector=value,
            )

    def get(self, entity: str):
        id = self._create_id(entity)
        resp = self.client.data_object.get_by_id(
            uuid=id, class_name=self.class_name, with_vector=True
        )
        return resp.get("vector")

    def nearest(
        self, feature: str, variant: str, vector: Union[List[float], np.ndarray], k: int
    ):
        if vector is np.ndarray:
            vector = vector.tolist()
        resp = (
            self.client.query.get(class_name=self.class_name, properties=["entity"])
            .with_near_vector(
                {
                    "vector": vector,
                }
            )
            .with_limit(k)
            .do()
        )
        matches = resp["data"]["Get"][self.class_name]
        return [m["entity"] for m in matches]

    def _create_id(self, entity: str) -> str:
        return str(uuid.uuid5(uuid.NAMESPACE_DNS, entity))


class WeaviateOnlineStore(VectorStore):
    def __init__(self, config: WeaviateConfig):
        self.class_name_template = "Featureform_class__{0}"
        self.client = weaviate.Client(
            url=config.url,
            auth_client_secret=weaviate.AuthApiKey(api_key=config.api_key),
        )

    def get_table(self, feature: str, variant: str):
        class_name = self._create_class_name(feature, variant)
        return WeaviateOnlineTable(
            client=self.client,
            class_name=class_name,
            vector_type=VectorType(
                scalar_type=ScalarType.FLOAT32,
                # Weaviate indexes (i.e. schema + classes) don't actually need the dimension upon
                # creation, and there doesn't appear to be a place to fetch the dimension from
                # the schema/class once it's been created; technically, in the context of localmode,
                # we don't need the dimension either; however, if it becomes useful, we'll need to
                # this and determine how to handle this. One option is to store the dimension
                # as a property of the class, but that would put the property field on every record
                # in the index.
                dimension=0,
                is_embedding=True,
            ),
        )

    def create_table(self, feature: str, variant: str, value_type: ValueType):
        return WeaviateOnlineTable(
            client=self.client,
            class_name=self._create_class_name(feature, variant),
            vector_type=value_type,
        )

    def delete_table(self, feature: str, variant: str):
        pass

    def table_exists(self, feature: str, variant: str) -> bool:
        class_name = self._create_class_name(feature, variant)
        return self.client.schema.exists(class_name)

    def close(self):
        pass

    def create_index(self, feature: str, variant: str, vector_type: VectorType):
        class_name = self._create_class_name(feature, variant)
        self.client.schema.create_class(
            {
                "class": class_name,
                "properties": [
                    {
                        "name": "entity",
                        "description": "The entity value (i.e. name/ID) of the feature variant",
                        "dataType": ["text"],
                        "indexInverted": False,
                    },
                ],
            }
        )
        return WeaviateOnlineTable(
            client=self.client,
            class_name=class_name,
            vector_type=vector_type,
        )

    def delete_index(self, feature: str, variant: str):
        class_name = self._create_class_name(feature, variant)
        self.client.schema.delete_class(class_name)

    def _create_class_name(self, feature: str, variant: str) -> str:
        # Class names must adhere to the following regex: /^[A-Z][_0-9A-Za-z]*$/
        # To ensure uniqueness and keep things simple for the user,
        # we created a deterministic UUID5 based on the feature and variant
        # and replace all "-" with "_" to ensure it adheres to the regex
        # See details: https://weaviate.io/developers/weaviate/config-refs/schema#class
        index_uuid = uuid.uuid5(uuid.NAMESPACE_DNS, f"{feature}-{variant}")
        index_uuid = str(index_uuid).replace("-", "_")
        return self.class_name_template.format(index_uuid)



import featureform as ff
from featureform import local
import pandas as pd
import pytest


def setup():
    transactions = local.register_file(
        name="transactions",
        variant="quickstart",
        description="A dataset of fraudulent transactions",
        path="transactions.csv",
    )

    @local.df_transformation(
        variant="quickstart", inputs=[("transactions", "quickstart")]
    )
    def average_user_transaction(transactions):
        """the average transaction amount for a user"""
        return transactions.groupby("CustomerID")["TransactionAmount"].mean()

    user = ff.register_entity("user")

    # Register a column from our transformation as a feature
    average_user_transaction.register_resources(
        entity=user,
        entity_column="CustomerID",
        inference_store=local,
        features=[
            {
                "name": "avg_transactions",
                "variant": "quickstart",
                "column": "TransactionAmount",
                "type": "float32",
            },
        ],
    )

    # Register label from our base Transactions table
    transactions.register_resources(
        entity=user,
        entity_column="CustomerID",
        timestamp_column="Timestamp",
        labels=[
            {
                "name": "fraudulent",
                "variant": "quickstart",
                "column": "IsFraud",
                "type": "bool",
            },
        ],
    )

    # Register label from our base Transactions table
    transactions.register_resources(
        entity=user,
        entity_column="CustomerID",
        labels=[
            {
                "name": "fraudulent",
                "variant": "quickstart-no-ts",
                "column": "IsFraud",
                "type": "bool",
            },
        ],
    )

    ff.register_training_set(
        "fraud_training",
        "quickstart",
        label=("fraudulent", "quickstart"),
        features=[("avg_transactions", "quickstart")],
    )

    ff.register_training_set(
        "fraud_training",
        "quickstart-no-ts",
        label=("fraudulent", "quickstart-no-ts"),
        features=[("avg_transactions", "quickstart")],
    )

    resource_client = ff.ResourceClient(local=True)
    resource_client.apply()


@pytest.mark.parametrize(
    "training_set_variant, include_label_timestamp, expected_in_keys",
    [
        ("quickstart", True, True),
        ("quickstart", False, False),
        ("quickstart-no-ts", True, False),
        ("quickstart-no-ts", False, False),
    ],
)
def test_include_label_timestamp(
    training_set_variant, include_label_timestamp, expected_in_keys
):
    serving_client = ff.ServingClient(local=True)
    dataset = serving_client.training_set(
        "fraud_training",
        training_set_variant,
        include_label_timestamp=include_label_timestamp,
    )
    dataset_pandas = dataset.pandas()

    result = "label_timestamp" in dataset_pandas.keys()

    assert result == expected_in_keys



import dill
import time

import pytest
import numpy as np

import featureform as ff
from featureform.resources import ResourceStatus, OnDemandFeatureVariant


@pytest.fixture(autouse=True)
def before_and_after_each(setup_teardown):
    setup_teardown()
    yield
    setup_teardown()


@pytest.mark.local
def test_ondemand_feature_decorator_class():
    name = "test_ondemand_feature"
    owner = "ff_tester"

    decorator = OnDemandFeatureVariant(owner=owner, name=name, variant="default")
    decorator_2 = OnDemandFeatureVariant(owner=owner, name=name, variant="default")

    assert decorator.name_variant() == (name, "default")
    assert decorator.type() == "ondemand_feature"
    assert decorator.get_status() == ResourceStatus.READY
    assert decorator.is_ready() is True
    assert decorator == decorator_2


@pytest.mark.local
def test_ondemand_decorator():
    owner = "ff_tester"

    @OnDemandFeatureVariant(owner=owner, variant="default")
    def test_fn():
        return 1 + 1

    expected_query = dill.dumps(test_fn.__code__)

    assert test_fn.name_variant() == (test_fn.__name__, "default")
    assert test_fn.query == expected_query


@pytest.mark.local
@pytest.mark.parametrize(
    # fmt: off
    "features,entity,expected_output",
    [
        (
                [
                    ("avg_transactions", "quickstart")
                ],
                {"user": "C8837983"},
                [1875.0]
        ),
        (
                [
                    ("pi", "default")
                ],
                {"user": "C8837983"},
                [3.141592653589793]
        ),
        (
                [
                    ("avg_transactions", "quickstart"),
                    ("pi", "default")
                ],
                {"user": "C8837983"},
                [1875.0, 3.141592653589793]
        ),
        (
                [
                    ("pi", "default"),
                    ("avg_transactions", "quickstart")
                ],
                {"user": "C8837983"},
                [3.141592653589793, 1875.0]
        ),
        (
                [
                    ("avg_transactions", "quickstart"),
                    ("pi", "default"),
                    ("avg_transactions", "quickstart")
                ],
                {"user": "C8837983"},
                [1875.0, 3.141592653589793, 1875.0]
        ),
        (
                [
                    ("avg_transactions", "quickstart"),
                    ("pi", "default"),
                    ("avg_transactions", "quickstart"),
                    ("pi", "default")
                ],
                {"user": "C8837983"},
                [1875.0, 3.141592653589793, 1875.0, 3.141592653589793]
        ),
        (
                [
                    ("pi", "default"),
                    ("pi", "pi_named"),
                    ("pi_called", "default")
                ],
                {},
                [3.141592653589793, 3.141592653589793, 3.141592653589793]
        ),
        # fmt: on
        pytest.param([], {}, [], marks=pytest.mark.xfail),
        pytest.param([("pi", "default")], None, [], marks=pytest.mark.xfail),
    ],
)
def test_serving_ondemand_precalculated_feature(features, entity, expected_output):
    register_resources()
    client = ff.ServingClient(local=True)

    features = client.features(features, entity)
    assert features.tolist() == expected_output

    client.impl.db.close()  # TODO automatically do this


def register_resources():
    ff.register_user("featureformer").make_default_owner()
    ff.set_run("default")
    local = ff.register_local()

    transactions = local.register_file(
        name="transactions",
        variant="quickstart",
        description="A dataset of fraudulent transactions",
        path="transactions.csv",
    )

    @local.df_transformation(
        variant="quickstart", inputs=[("transactions", "quickstart")]
    )
    def average_user_transaction(transactions):
        """the average transaction amount for a user"""
        return transactions.groupby("CustomerID")["TransactionAmount"].mean()

    @ff.ondemand_feature
    def pi(serving_client, entities, params):
        import math

        return math.pi

    @local.ondemand_feature()
    def pi_called(serving_client, entities, params):
        import math

        return math.pi

    @ff.ondemand_feature(variant="pi_named")
    def pi(serving_client, entities, params):
        import math

        return math.pi

    user = ff.register_entity("user")
    # Register a column from our transformation as a feature
    average_user_transaction.register_resources(
        entity=user,
        entity_column="CustomerID",
        inference_store=local,
        features=[
            {
                "name": "avg_transactions",
                "variant": "quickstart",
                "column": "TransactionAmount",
                "type": "float32",
            },
        ],
    )
    # Register label from our base Transactions table
    transactions.register_resources(
        entity=user,
        entity_column="CustomerID",
        labels=[
            {
                "name": "fraudulent",
                "variant": "quickstart",
                "column": "IsFraud",
                "type": "bool",
            },
        ],
    )

    ff.register_training_set(
        "fraud_training",
        "quickstart",
        label=("fraudulent", "quickstart"),
        features=[("avg_transactions", "quickstart")],
    )

    resource_client = ff.ResourceClient(local=True)
    resource_client.apply()



from typing import List, Any, Union
import uuid
import pinecone
import numpy as np

from .online_store import VectorStoreTable, VectorStore, ValueType, VectorType
from ..resources import PineconeConfig
from ..enums import ScalarType


class PineconeOnlineTable(VectorStoreTable):
    def __init__(self, client, index_name: str, namespace: str, vector_type: ValueType):
        self.client = client
        self.index_name = index_name
        self.namespace = namespace
        self.vector_type = vector_type

    def set(self, entity: str, value: Union[List[float], np.ndarray]):
        self.client.Index(self.index_name).upsert(
            vectors=[(self._create_id(entity), value, {"id": entity})],
            namespace=self.namespace,
        )

    def get(self, entity: str):
        id = self._create_id(entity)
        fetch_response = self.client.Index(self.index_name).fetch(
            ids=[id],
            namespace=self.namespace,
        )
        vector = fetch_response["vectors"].get(id)
        if vector is None:
            return None
        else:
            return vector["values"]

    def nearest(
        self, feature: str, variant: str, vector: Union[List[float], np.ndarray], k: int
    ):
        name_variant_id = uuid.uuid5(uuid.NAMESPACE_DNS, f"{feature}-{variant}")
        if isinstance(vector, np.ndarray):
            vector = vector.tolist()
        query_response = self.client.Index(f"ff-idx--{name_variant_id}").query(
            namespace=self.namespace, vector=vector, top_k=k, include_metadata=True
        )
        matches = query_response.get("matches", [])
        if len(matches) == 0:
            return []
        else:
            return [match["metadata"]["id"] for match in matches]

    def _create_id(self, entity: str) -> str:
        id_uuid = uuid.uuid5(uuid.NAMESPACE_DNS, entity)
        return str(id_uuid)


class PineconeOnlineStore(VectorStore):
    def __init__(self, config: PineconeConfig):
        self.index_name_template = "ff-idx--{0}"
        pinecone.init(api_key=config.api_key, environment=config.environment)
        self.client = pinecone

    def get_table(self, feature: str, variant: str) -> PineconeOnlineTable:
        idx = self._create_index_name(feature, variant)
        describe_response = self.client.Index(idx).describe_index_stats()
        return PineconeOnlineTable(
            client=self.client,
            index_name=idx,
            namespace=self._create_namespace(feature, variant),
            vector_type=VectorType(
                scalar_type=ScalarType.FLOAT32,
                dimension=describe_response.dimension,
                is_embedding=True,
            ),
        )

    def create_table(self, feature: str, variant: str, entity_type: ValueType):
        return PineconeOnlineTable(
            client=self.client,
            index_name=self._create_index_name(feature, variant),
            namespace=self._create_namespace(feature, variant),
            vector_type=entity_type,
        )

    def table_exists(self, feature: str, variant: str) -> bool:
        try:
            self.client.describe_index(self._create_index_name(feature, variant))
            return True
        except pinecone.core.client.exceptions.NotFoundException:
            return False

    def delete_index(self, feature: str, variant: str):
        index_name = self._create_index_name(feature, variant)
        self.client.delete_index(name=index_name)

    def delete_table(self, feature: str, variant: str):
        pass

    def close(self):
        pass

    def create_index(self, feature: str, variant: str, vector_type: VectorType):
        index_name = self._create_index_name(feature, variant)
        self.client.create_index(
            name=index_name, dimension=vector_type.dimension, metric="cosine"
        )
        return PineconeOnlineTable(
            client=self.client,
            index_name=self._create_index_name(feature, variant),
            namespace=self._create_namespace(feature, variant),
            vector_type=vector_type,
        )

    def _create_index_name(self, feature: str, variant: str) -> str:
        index_uuid = uuid.uuid5(uuid.NAMESPACE_DNS, f"{feature}-{variant}")
        return self.index_name_template.format(index_uuid)

    def _create_namespace(self, feature: str, variant: str) -> str:
        return f"ff-namespace--{feature}-{variant}"



import json
import os
from functools import lru_cache
from typing import Callable, Set

import pandas as pd
from featureform import SQLiteMetadata
from featureform.local_utils import get_sql_transformation_sources
from featureform.resources import SourceType  # fix to do client.source.import
from pandas.core.generic import NDFrame
from typeguard import typechecked

from .file_utils import absolute_file_paths


class LocalCache:
    def __init__(self):
        feature_form_dir = os.environ.get("FEATUREFORM_DIR", ".featureform")
        self.cache_dir = os.environ.get(
            "FEATUREFORM_CACHE_DIR", os.path.join(feature_form_dir, "cache")
        )

    @typechecked
    def get_or_put(
        self,
        resource_type: str,
        resource_name: str,
        resource_variant: str,
        source_name: str,
        source_variant: str,
        func: Callable[[], NDFrame],
    ) -> NDFrame:
        """
        Caches the result of a callable to a local file. If the source files have changed, the cache is invalidated.
        """
        cache_file_path = self._cache_file_path(
            resource_type, resource_name, resource_variant
        )

        with SQLiteMetadata() as db:
            # check db for source files
            source_files_from_db = db.get_source_files_for_resource(
                resource_type, resource_name, resource_variant
            )
            if source_files_from_db:
                self._invalidate_cache_if_source_files_changed(
                    source_files_from_db, cache_file_path
                )

            # get source files from db or compute the sources
            source_files: Set[str] = (
                set(map(lambda x: x["file_path"], source_files_from_db))
                if source_files_from_db
                else self.get_source_files_for_source(db, source_name, source_variant)
            )

            return self._get_or_put(
                db,
                resource_type,
                resource_name,
                resource_variant,
                cache_file_path,
                source_files,
                func,
            )

    @typechecked
    def get_or_put_training_set(
        self,
        training_set_name: str,
        training_set_variant: str,
        func: Callable[[], NDFrame],
    ) -> NDFrame:
        """
        Caches the result of a training set to a local file. Difference between this one and the one above
        is how this needs to fetch all the source files for the training set.
        """
        resource_type = "training_set"

        file_path = self._cache_file_path(
            resource_type, training_set_name, training_set_variant
        )

        with SQLiteMetadata() as db:
            # check db for source files
            source_files_from_db = db.get_source_files_for_resource(
                resource_type, training_set_name, training_set_variant
            )

            # Only check to invalidate the cache if we have source files in the db
            if source_files_from_db:
                self._invalidate_cache_if_source_files_changed(
                    source_files_from_db, file_path
                )

            source_files = set()
            if source_files_from_db:
                source_files.update(
                    set(map(lambda x: x["file_path"], source_files_from_db))
                )
            else:
                ts_variant = db.get_training_set_variant(
                    training_set_name, training_set_variant
                )
                label_variant = db.get_label_variant(
                    ts_variant["label_name"], ts_variant["label_variant"]
                )
                source_files.update(
                    self.get_source_files_for_source(
                        db,
                        label_variant["source_name"],
                        label_variant["source_variant"],
                    )
                )

                features = db.get_training_set_features(
                    training_set_name, training_set_variant
                )
                for feature in features:
                    feature_variant = db.get_feature_variant(
                        feature["feature_name"], feature["feature_variant"]
                    )
                    source_files.update(
                        self.get_source_files_for_source(
                            db,
                            feature_variant["source_name"],
                            feature_variant["source_variant"],
                        )
                    )

            return self._get_or_put(
                db,
                resource_type,
                training_set_name,
                training_set_variant,
                file_path,
                source_files,
                func,
            )

    def _get_or_put(
        self,
        db,
        resource_type,
        resource_name,
        resource_variant,
        file_path,
        source_files,
        func,
    ) -> NDFrame:
        if os.path.exists(file_path):
            return pd.read_pickle(file_path)
        else:
            # create the dir if not exists and write the file
            df = func()
            os.makedirs(self.cache_dir, exist_ok=True)
            df.to_pickle(file_path)
            for source_file in source_files:
                db.insert_or_update(
                    "resource_source_files",
                    ["resource_type", "name", "variant", "file_path"],
                    ["updated_at"],
                    resource_type,
                    resource_name,
                    resource_variant,
                    source_file,
                    str(os.path.getmtime(source_file)),
                )
            return df

    @lru_cache(maxsize=128)
    def get_source_files_for_source(self, db, source_name, source_variant) -> Set[str]:
        """
        Recursively gets the source files for a given source. Each call is cached.
        """
        source = db.get_source_variant(source_name, source_variant)
        transform_type = db.is_transformation(source_name, source_variant)

        sources = set()
        if transform_type == SourceType.PRIMARY_SOURCE.value:
            return {source["definition"]}
        elif transform_type == SourceType.SQL_TRANSFORMATION.value:
            query = source["definition"]
            transformation_sources = get_sql_transformation_sources(query)
            for source_name, source_variant in transformation_sources:
                sources.update(
                    self.get_source_files_for_source(db, source_name, source_variant)
                )
        elif transform_type == SourceType.DF_TRANSFORMATION.value:
            dependencies = json.loads(source["inputs"])
            for name, variant in dependencies:
                sources.update(self.get_source_files_for_source(db, name, variant))
        elif transform_type == SourceType.DIRECTORY.value:
            path = source["definition"]
            for absolute_file, _ in absolute_file_paths(path):
                sources.add(absolute_file)
        else:
            raise Exception(f"Unknown source type: {transform_type}")
        return sources

    def _invalidate_cache_if_source_files_changed(
        self, source_files_from_db, cache_file_path
    ):
        if any(
            self._file_has_changed(source_file["updated_at"], source_file["file_path"])
            for source_file in source_files_from_db
        ):
            if os.path.exists(cache_file_path):
                os.remove(cache_file_path)

    def _cache_file_path(self, resource_type: str, name: str, variant: str):
        key = f"{resource_type}__{name}__{variant}"
        return f"{self.cache_dir}/{key}.pkl"

    @staticmethod
    def _file_has_changed(last_updated_at, file_path):
        """
        Currently using last updated at for determining if a file has changed. We can consider using the file hash
        if this becomes a performance issue.
        """
        os_last_updated = os.path.getmtime(file_path)
        return os_last_updated > float(last_updated_at)



from .filestore import LocalFileStore
from .pinecone import PineconeOnlineStore
from .weaviate import WeaviateOnlineStore
from .online_store import OnlineStore


def get_provider(provider) -> OnlineStore:
    providers = {
        "LOCAL_ONLINE": LocalFileStore,
        "PINECONE_ONLINE": PineconeOnlineStore,
        "WEAVIATE_ONLINE": WeaviateOnlineStore,
    }
    if provider in providers:
        return providers[provider]
    else:
        raise NotImplementedError(f"Provider {provider} not implemented")



import json
import os
import sqlite3
from threading import Lock

from .exceptions import FeatureNotFound


class SyncSQLExecutor:
    def __init__(self, conn):
        self.__conn = conn
        self.__lock = Lock()

    def execute(self, cmd):
        with self.__lock:
            return self.__conn.execute(cmd)

    def execute_stmt(self, stmt, vals):
        with self.__lock:
            return self.__conn.execute(stmt, vals)

    def executemany(self, cmd, param):
        with self.__lock:
            return self.__conn.executemany(cmd, param)

    def close(self):
        with self.__lock:
            return self.__conn.close()

    def commit(self):
        with self.__lock:
            return self.__conn.commit()


class SQLiteMetadata:
    def __init__(self, path=".featureform/SQLiteDB"):
        self.path = path
        if not os.path.exists(self.path):
            os.makedirs(self.path)
        raw_conn = sqlite3.connect(f"{self.path}/metadata.db", check_same_thread=False)
        raw_conn.row_factory = sqlite3.Row
        self.__conn = SyncSQLExecutor(raw_conn)
        self.createTables()

    def createTables(self):
        # Features variant table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS feature_variant(
          created text,
          description text,
          entity text NOT NULL,
          name text NOT NULL,
          owner text,
          provider text NOT NULL,
          data_type text NOT NULL,
          variant text NOT NULL,
          status text,
          source_entity text,
          source_timestamp text,
          source_value text,
          source_name text NOT NULL,
          source_variant text NOT NULL,
          is_embedding bool, 
          dimension int,

          PRIMARY KEY(name, variant),

          FOREIGN KEY(name) REFERENCES features(name),
          FOREIGN KEY(entity) REFERENCES entities(name),
          FOREIGN KEY(provider) REFERENCES providers(name),
          FOREIGN KEY(source_name) REFERENCES sources(name))"""
        )

        # OnDemand Features variant table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS ondemand_feature_variant(
          created text,
          description text,
          name text NOT NULL,
          owner text,
          variant text NOT NULL,
          status text,
          query text,

          PRIMARY KEY(name, variant),

          FOREIGN KEY(name) REFERENCES features(name))"""
        )

        # Features table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS features(
          name text NOT NULL,
          default_variant text NOT NULL,
          type text,
          PRIMARY KEY (name));"""
        )

        # features type table
        # this is used to determine if it is a pre-calculated or client-calculated feature
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS feature_computation_mode (
          name text NOT NULL,
          variant text NOT NULL,
          mode text,
          is_on_demand integer,
          PRIMARY KEY (name, variant));"""
        )

        # training set variant
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS training_set_variant(
          created text,
          description text,            
          name text NOT NULL,
          owner text,
          variant text,
          label_name text,
          label_variant text,
          status text,
          PRIMARY KEY(name, variant),
          FOREIGN KEY(name) REFERENCES training_sets(name));"""
        )

        # Training-set table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS training_sets(
          type text NOT NULL,
          default_variant text,
          name text PRIMARY KEY NOT NULL);"""
        )

        # Training set features
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS training_set_features(
          training_set_name text NOT NULL,
          training_set_variant text NOT NULL,
          feature_name text NOT NULL,
          feature_variant text NOT NULL,
          UNIQUE(training_set_name, training_set_variant, feature_name, feature_variant));"""
        )

        # Training set lag features
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS training_set_lag_features(
          training_set_name text NOT NULL,
          training_set_variant text NOT NULL,
          feature_name text NOT NULL,
          feature_variant text NOT NULL,
          feature_new_name text,
          feature_lag real,
          UNIQUE(training_set_name, training_set_variant, feature_name, feature_variant, feature_new_name, feature_lag));"""
        )

        # source variant
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS source_variant(
          created     text,
          description text,
          name  text NOT NULL,
          source_type  text,
          owner       text,
          provider    text NOT NULL,
          variant    text,
          status      text,
          transformation text,
          inputs text, 
          definition  BLOB,
          PRIMARY KEY(name, variant),
          FOREIGN KEY(provider) REFERENCES providers(name),
          FOREIGN KEY(name) REFERENCES sources(name));"""
        )

        # source variant text
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS source_variant_text(
            created     text,
            source_name  text NOT NULL,
            variant     text NOT NULL,
            source_text text,
            PRIMARY KEY(source_name, variant));"""
        )

        # sources table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS sources(
          type           text NOT NULL,
          default_variant text,
          name           text PRIMARY KEY NOT NULL);"""
        )

        # labels variant
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS label_variant(
          created         text,
          description     text,
          entity          text,
          name            text NOT NULL,
          owner           text,
          provider        text,
          data_type       text,
          variant         text,
          source_entity    text,
          source_timestamp text,
          source_value     text,
          status          text,
          source_name      text,
          source_variant   text,
          FOREIGN KEY(provider) REFERENCES providers(name),
          PRIMARY KEY(name, variant),
          FOREIGN KEY(name) REFERENCES labels(name));"""
        )

        # labels table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS labels(
          type           text,
          default_variant text,
          name           text PRIMARY KEY);"""
        )

        # entity table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS entities(
          name        text PRIMARY KEY NOT NULL,
          type        text,
          description text,
          status      text);"""
        )

        # user table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS users(
          name   text PRIMARY KEY NOT NULL,
          type   text,
          status text);"""
        )

        # models table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS models(
          name        text PRIMARY KEY NOT NULL,
          type        text);"""
        )

        # models training sets table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS model_training_sets(
          model_name           text NOT NULL,
          training_set_name    text NOT NULL,
          training_set_variant text NOT NULL,
          UNIQUE(model_name, training_set_name, training_set_variant));"""
        )

        # models features sets table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS model_features(
          model_name           text NOT NULL,
          feature_name    text NOT NULL,
          feature_variant text NOT NULL,
          UNIQUE(model_name, feature_name, feature_variant));"""
        )

        # providers table
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS providers(
          name             text PRIMARY KEY NOT NULL,
          type             text,
          description      text,
          provider_type     text,
          software         text,
          team             text,
          sources          text,
          status           text,
          serialized_config text)"""
        )

        # source files for a resource
        # tracks the source files that have been used to create a resource and contains the files last_updated
        # this is used to determined if caches need to be invalidated
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS resource_source_files(
            resource_type text NOT NULL,
            name text NOT NULL,
            variant text NOT NULL,
            file_path text NOT NULL,
            updated_at text NOT NULL,
            PRIMARY KEY(resource_type, name, variant, file_path))"""
        )

        # full-text search table
        self.__conn.execute(
            """CREATE VIRTUAL TABLE IF NOT EXISTS resources_fts USING fts5(
          resource_type, -- an "enum" column to filter with (e.g. feature, training_set, source, provider, entity, label, model, user)
          name,
          variant,
          description,
          status,
        );"""
        )

        # resource table triggers to populate fts table
        self.__conn.execute(
            """CREATE TRIGGER IF NOT EXISTS feature_variant_after_insert
        AFTER INSERT ON feature_variant
        BEGIN
        INSERT INTO resources_fts(resource_type, name, variant, description, status)
        VALUES ('feature', new.name, new.variant, new.description, new.status);
        END;"""
        )

        self.__conn.execute(
            """CREATE TRIGGER IF NOT EXISTS training_set_variant_after_insert
        AFTER INSERT ON training_set_variant
        BEGIN
        INSERT INTO resources_fts(resource_type, name, variant, description, status)
        VALUES ('training_set', new.name, new.variant, new.description, new.status);
        END;"""
        )
        self.__conn.execute(
            """CREATE TRIGGER IF NOT EXISTS source_variant_after_insert
        AFTER INSERT ON source_variant
        BEGIN
        INSERT INTO resources_fts(resource_type, name, variant, description, status)
        VALUES ('source', new.name, new.variant, new.description, new.status);
        END;"""
        )

        self.__conn.execute(
            """CREATE TRIGGER IF NOT EXISTS label_variant_after_insert
        AFTER INSERT ON label_variant
        BEGIN
        INSERT INTO resources_fts(resource_type, name, variant, description, status)
        VALUES ('label', new.name, new.variant, new.description, new.status);
        END;"""
        )

        self.__conn.execute(
            """CREATE TRIGGER IF NOT EXISTS entities_after_insert
        AFTER INSERT ON entities
        BEGIN
        INSERT INTO resources_fts(resource_type, name, variant, description, status)
        VALUES ('entity', new.name, '', new.description, new.status);
        END;"""
        )

        self.__conn.execute(
            """CREATE TRIGGER IF NOT EXISTS users_after_insert
        AFTER INSERT ON users
        BEGIN
        INSERT INTO resources_fts(resource_type, name, variant, description, status)
        VALUES ('user', new.name, '', '', new.status);
        END;"""
        )

        self.__conn.execute(
            """CREATE TRIGGER IF NOT EXISTS models_after_insert
        AFTER INSERT ON models
        BEGIN
        INSERT INTO resources_fts(resource_type, name, variant, description, status)
        VALUES ('model', new.name, '', '', '');
        END;"""
        )

        self.__conn.execute(
            """CREATE TRIGGER IF NOT EXISTS provider_after_insert
        AFTER INSERT ON providers
        BEGIN
        INSERT INTO resources_fts(resource_type, name, variant, description, status)
        VALUES ('provider', new.name, '', new.description, new.status);
        END;"""
        )

        # Tags and Properties tables
        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS tags(
          name text NOT NULL,
          variant text,
          type text,
          tag_list text,
          UNIQUE(name,variant,type));"""
        )

        self.__conn.execute(
            """CREATE TABLE IF NOT EXISTS properties(
          name text NOT NULL,
          variant text,
          type text,
          property_list text,
          UNIQUE(name,variant,type));"""
        )

        self.__conn.commit()

    def get_type_table(self, tablename):
        if tablename in ["entities", "models", "users", "providers"]:
            query = """SELECT r.*, t.tag_list as tags, p.property_list as properties
          FROM {0} r
          LEFT JOIN tags t ON t.name = r.name
          LEFT JOIN properties p ON p.name = r.name     ;
          """.format(
                tablename
            )
        else:
            query = f"SELECT * FROM {tablename}"
        type_data = self.__conn.execute(query)
        self.__conn.commit()
        return type_data.fetchall()

    def query_resource(
        self, table, column=None, resource_name=None, should_fetch_tags_properties=False
    ):
        if should_fetch_tags_properties:
            query = f"""SELECT r.*, t.tag_list as tags, p.property_list as properties
                FROM {table} r
                LEFT JOIN tags t ON t.name = r.name
                LEFT JOIN properties p ON p.name = r.name
           """
            if column is not None:
                query = query + f"WHERE r.{column} = '{resource_name}';"
        else:
            query = f"SELECT * FROM {table}"
            if column is not None:
                query = query + f" WHERE {column} = '{resource_name}'"
        resource_data = self.__conn.execute(query)
        self.__conn.commit()
        resource_data_list = resource_data.fetchall()
        if (len(resource_data_list)) == 0:
            return []
        return resource_data_list

    def query_resource_variant(self, table, column, resource_name):
        query = """SELECT r.*, t.tag_list as tags, p.property_list as properties
           FROM {0} r
           LEFT JOIN tags t ON t.name = r.name AND t.variant = r.variant
           LEFT JOIN properties p ON p.name = r.name AND p.variant = r.variant
           WHERE r.{1} = '{2}';""".format(
            table, column, resource_name
        )
        variant_data = self.__conn.execute(query)
        self.__conn.commit()
        variant_data_list = variant_data.fetchall()
        if len(variant_data_list) == 0 and column != "owner":
            raise ValueError(f"{table} with {column}: {resource_name} not found")
        return variant_data_list

    def fetch_data(self, query, type, name, variant):
        variant_data = self.__conn.execute(query)
        self.__conn.commit()
        variant_data_list = variant_data.fetchall()
        if len(variant_data_list) == 0:
            raise ValueError(f"{type} {name} ({variant}) not found")
        return variant_data_list

    def fetch_data_safe(self, query, type, name, variant):
        variant_data = self.__conn.execute(query)
        self.__conn.commit()
        variant_data_list = variant_data.fetchall()
        if len(variant_data_list) == 0:
            return []
        return variant_data_list

    def get_user(self, name, should_fetch_tags_properties):
        return self.query_resource("users", "name", name, should_fetch_tags_properties)[
            0
        ]

    def get_entity(self, name, should_fetch_tags_properties):
        return self.query_resource(
            "entities", "name", name, should_fetch_tags_properties
        )[0]

    def get_feature(self, name, should_fetch_tags_properties):
        return self.query_resource(
            "features", "name", name, should_fetch_tags_properties
        )[0]

    def get_label(self, name, should_fetch_tags_properties):
        return self.query_resource(
            "labels", "name", name, should_fetch_tags_properties
        )[0]

    def get_source(self, name, should_fetch_tags_properties):
        return self.query_resource(
            "sources", "name", name, should_fetch_tags_properties
        )[0]

    def get_training_set(self, name, should_fetch_tags_properties):
        return self.query_resource(
            "training_sets", "name", name, should_fetch_tags_properties
        )[0]

    def get_model(self, name, should_fetch_tags_properties):
        return self.query_resource(
            "models", "name", name, should_fetch_tags_properties
        )[0]

    def get_provider(self, name, should_fetch_tags_properties):
        return self.query_resource(
            "providers", "name", name, should_fetch_tags_properties
        )[0]

    def get_feature_variant(self, name, variant):
        query = """SELECT fv.*, t.tag_list as tags, p.property_list as properties
        FROM feature_variant fv
        LEFT JOIN tags t ON t.name = fv.name AND t.variant = fv.variant
        LEFT JOIN properties p ON p.name = fv.name AND p.variant = fv.variant
        WHERE fv.name = '{0}' AND fv.variant = '{1}';""".format(
            name, variant
        )
        return self.fetch_data(query, "feature_variant", name, variant)[0]

    def get_feature_variants_from_provider(self, name):
        return self.query_resource_variant("feature_variant", "provider", name)

    def get_feature_variants_from_source(self, name, variant):
        query = """SELECT fv.*, t.tag_list as tags, p.property_list as properties
        FROM feature_variant fv
        LEFT JOIN tags t ON t.name = fv.name AND t.variant = fv.variant
        LEFT JOIN properties p ON p.name = fv.name AND p.variant = fv.variant
        WHERE fv.source_name = '{0}' AND fv.source_variant = '{1}';""".format(
            name, variant
        )
        return self.fetch_data_safe(query, "feature_variant", name, variant)

    def get_feature_variants_from_feature(self, name):
        return self.query_resource_variant("feature_variant", "name", name)

    def get_feature_variant_mode(self, name, variant):
        query = f"SELECT mode FROM feature_computation_mode WHERE name='{name}' AND variant='{variant}'"

        feature_metadata = self.fetch_data_safe(
            query, "feature_computation_mode", name, variant
        )

        if len(feature_metadata) == 0:
            raise FeatureNotFound(name, variant)

        return feature_metadata[0]["mode"]

    def get_feature_variant_on_demand(self, name, variant):
        query = f"SELECT is_on_demand FROM feature_computation_mode WHERE name='{name}' AND variant='{variant}'"
        return bool(
            self.fetch_data_safe(query, "feature_computation_mode", name, variant)[0][
                "is_on_demand"
            ]
        )

    def get_ondemand_feature_query(self, name, variant):
        query = f"SELECT query FROM ondemand_feature_variant WHERE name='{name}' AND variant='{variant}'"
        return self.fetch_data_safe(query, "ondemand_feature_variant", name, variant)[
            0
        ]["query"]

    def get_training_set_variant(self, name, variant):
        query = """SELECT v.*, t.tag_list as tags, p.property_list as properties
        FROM training_set_variant v
        LEFT JOIN tags t ON t.name = v.name AND t.variant = v.variant
        LEFT JOIN properties p ON p.name = v.name AND p.variant = v.variant
        WHERE v.name = '{0}' AND v.variant = '{1}';
        """.format(
            name, variant
        )
        return self.fetch_data(query, "training_set_variant", name, variant)[0]

    def get_training_set_variant_from_training_set(self, name):
        return self.query_resource_variant("training_set_variant", "name", name)

    def get_training_set_variant_from_label(self, name, variant):
        query = """SELECT v.*, t.tag_list as tags, p.property_list as properties
        FROM training_set_variant v
        LEFT JOIN tags t ON t.name = v.name AND t.variant = v.variant
        LEFT JOIN properties p ON p.name = v.name AND p.variant = v.variant
        WHERE v.label_name = '{0}' AND v.label_variant = '{1}';
        """.format(
            name, variant
        )
        return self.fetch_data_safe(query, "training_set_variant", name, variant)

    def get_label_variant(self, name, variant):
        query = """SELECT v.*, t.tag_list as tags, p.property_list as properties
        FROM label_variant v
        LEFT JOIN tags t ON t.name = v.name AND t.variant = v.variant
        LEFT JOIN properties p ON p.name = v.name AND p.variant = v.variant
        WHERE v.name = '{0}' AND v.variant = '{1}';
        """.format(
            name, variant
        )
        return self.fetch_data(query, "label_variant", name, variant)[0]

    def get_label_variants_from_label(self, name):
        return self.query_resource_variant("label_variant", "name", name)

    def get_label_variants_from_provider(self, name):
        return self.query_resource_variant("label_variant", "provider", name)

    def get_label_variants_from_source(self, name, variant):
        query = """SELECT v.*, t.tag_list as tags, p.property_list as properties
        FROM label_variant v
        LEFT JOIN tags t ON t.name = v.name AND t.variant = v.variant
        LEFT JOIN properties p ON p.name = v.name AND p.variant = v.variant
        WHERE v.source_name = '{0}' AND v.source_variant = '{1}';
        """.format(
            name, variant
        )
        return self.fetch_data_safe(query, "label_variant", name, variant)

    def get_source_variant(self, name, variant):
        query = """SELECT v.*, t.tag_list as tags, p.property_list as properties
        FROM source_variant v
        LEFT JOIN tags t ON t.name = v.name AND t.variant = v.variant
        LEFT JOIN properties p ON p.name = v.name AND p.variant = v.variant
        WHERE v.name = '{0}' AND v.variant = '{1}';
        """.format(
            name, variant
        )
        return self.fetch_data(query, "source_variant", name, variant)[0]

    def get_source_variants_from_source(self, name):
        return self.query_resource_variant("source_variant", "name", name)

    def get_training_set_from_features(self, name, variant):
        query = """SELECT v.*, t.tag_list as tags, p.property_list as properties
        FROM training_set_features v
        LEFT JOIN tags t ON t.name = v.training_set_name AND t.variant = v.training_set_variant
        LEFT JOIN properties p ON p.name = v.training_set_name AND p.variant = v.training_set_variant
        WHERE v.feature_name = '{0}' AND v.feature_variant = '{1}';
        """.format(
            name, variant
        )
        return self.fetch_data_safe(query, "training_set_features", name, variant)

    def get_training_set_from_labels(self, name, variant):
        query = """SELECT v.*, t.tag_list as tags, p.property_list as properties
        FROM training_set_variant v
        LEFT JOIN tags t ON t.name = v.name AND t.variant = v.variant
        LEFT JOIN properties p ON p.name = v.name AND p.variant = v.variant
        WHERE v.label_name = '{0}' AND v.label_variant = '{1}';
        """.format(
            name, variant
        )
        return self.fetch_data_safe(query, "training_set_variant", name, variant)

    def get_training_set_features(self, name, variant):
        query = """SELECT v.*, t.tag_list as tags, p.property_list as properties
        FROM training_set_features v
        LEFT JOIN tags t ON t.name = v.training_set_name AND t.variant = v.training_set_variant
        LEFT JOIN properties p ON p.name = v.training_set_name AND p.variant = v.training_set_variant
         WHERE v.training_set_name = '{0}' AND v.training_set_variant = '{1}';
        """.format(
            name, variant
        )
        return self.fetch_data(query, "training_set_features", name, variant)

    def get_training_set_lag_features(self, name, variant):
        query = """SELECT v.*, t.tag_list as tags, p.property_list as properties
        FROM training_set_lag_features v
        LEFT JOIN tags t ON t.name = v.training_set_name AND t.variant = v.training_set_variant
        LEFT JOIN properties p ON p.name = v.training_set_name AND p.variant = v.training_set_variant
         WHERE v.training_set_name = '{0}' AND v.training_set_variant = '{1}';
        """.format(
            name, variant
        )
        return self.fetch_data_safe(query, "training_set_lag_features", name, variant)

    def get_resource_with_source(self, tablename, source_name, source_variant):
        query = """SELECT v.*, t.tag_list as tags, p.property_list as properties
        FROM {0} v
        LEFT JOIN tags t ON t.name = v.name AND t.variant = v.variant
        LEFT JOIN properties p ON p.name = v.name AND p.variant = v.variant
         WHERE v.source_name = '{1}' AND v.source_variant = '{2}';
        """.format(
            tablename, source_name, source_variant
        )
        return self.fetch_data(query, tablename, source_name, source_variant)

    def is_transformation(self, name, variant):
        query = f"SELECT transformation FROM source_variant WHERE name='{name}' and variant='{variant}';"
        transformation = self.__conn.execute(query)
        self.__conn.commit()
        t = transformation.fetchall()
        if len(t) == 0:
            return 0
        return t[0][0]

    def get_source_files_for_resource(self, resource_type, name, variant):
        query = f"SELECT * FROM resource_source_files WHERE resource_type='{resource_type}' and name='{name}' and variant='{variant}';"
        result = self.__conn.execute(query)
        self.__conn.commit()
        return result.fetchall()

    def get_source_file_last_updated(self, resource_type, name, variant, file_path):
        query = f"SELECT * FROM resource_source_files WHERE resource_type='{resource_type}' and name='{name}' and variant='{variant}' and file_path='{file_path}';"
        result = self.__conn.execute(query)
        self.__conn.commit()
        res = result.fetchone()
        return res[0] if res else None

    def insert_source(self, tablename, *args):
        stmt = f"INSERT OR IGNORE INTO {tablename} VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"
        self.__conn.execute_stmt(stmt, args)
        self.__conn.commit()

    def insert_source_variant_text(self, *args):
        stmt = f"INSERT OR IGNORE INTO source_variant_text VALUES (?, ?, ?, ?)"
        self.__conn.execute_stmt(stmt, args)
        self.__conn.commit()

    def get_source_variant_text(self, name, variant):
        stmt = f"SELECT source_text FROM source_variant_text WHERE source_name='{name}' AND variant='{variant}'"
        result = self.__conn.execute(stmt)
        self.__conn.commit()
        res = result.fetchone()
        return res[0] if res else ""

    def insert(self, tablename, *args):
        query = f"INSERT OR IGNORE INTO {tablename} VALUES {str(args)}"
        self.__conn.execute(query)
        self.__conn.commit()

    def get_tags(self, name, variant, resource_type):
        query = f"SELECT tag_list FROM tags WHERE name='{name}' and variant='{variant}' and type='{resource_type}'"
        return self.fetch_data_safe(query, resource_type, name, variant)

    def update_tags(self, name, variant, type, tagList):
        query = f"UPDATE tags SET tag_list='{tagList}' WHERE name='{name}' and variant='{variant}' and type='{type}';"
        self.__conn.execute(query)
        self.__conn.commit()

    def insert_or_update(self, tablename, keys, cols, *args):
        """
        Upserts a row into the table. `keys` indicate columns that are unique
        and `cols` are the columns that are updated.
        """
        query = (
            f"INSERT INTO {tablename} VALUES {str(args)} "
            f"ON CONFLICT ({','.join(keys)}) DO UPDATE SET {','.join([f'{col}=excluded.{col}' for col in cols])}"
        )
        self.__conn.execute(query)
        self.__conn.commit()

    # TODO get something generic working, possibly consider using the above insert or update
    def upsert(self, tablename, *args):
        if tablename == "tags" or tablename == "properties":
            query, is_update = self.upsert_tags_properties(tablename, *args)
        else:
            raise NotImplementedError(f"UPSERT not implemented for table {tablename}")

        if is_update:
            self.__conn.execute(query)
            self.__conn.commit()
        else:
            self.insert(tablename, *args)

    def close(self):
        self.__conn.close()

    def search(self, phrase, resource_type=None):
        if resource_type is None:
            query = f"SELECT * FROM resources_fts WHERE resources_fts match '{{name variant description status}} : {phrase}*';"
        else:
            query = f"SELECT * FROM resources_fts WHERE resources_fts match '{{name variant description status}} : {phrase}*' AND resource_type = '{resource_type}';"

        search_results = self.__conn.execute(query)
        self.__conn.commit()
        results_list = search_results.fetchall()

        return results_list

    def upsert_tags_properties(self, tablename, *args):
        result_idx = 0
        type_idx = 2
        updatable_column_idx = -1
        query = ""
        updatable_column = "tag_list" if tablename == "tags" else "property_list"
        default_value = [] if tablename == "tags" else {}
        is_update = False
        name = args[0]
        variant = args[1]

        cursor = self.__conn.execute(
            f"SELECT * FROM {tablename} WHERE name = '{name}' AND variant = '{variant}';"
        )
        self.__conn.commit()
        existing = cursor.fetchall()

        if len(existing):
            existing_val = (
                json.loads(existing[result_idx][updatable_column])
                if existing[result_idx][updatable_column] is not None
                else default_value
            )
            if tablename == "tags":
                updated = json.dumps(
                    list(set(existing_val + json.loads(args[updatable_column_idx])))
                )
            else:
                updated = json.dumps(
                    {**existing_val, **json.loads(args[updatable_column_idx])}
                )
            args = list(args)
            args[updatable_column_idx] = updated
            query = f"UPDATE {tablename} SET name = '{name}', variant = '{variant}', type = '{args[type_idx]}', {updatable_column} = '{args[updatable_column_idx]}'"
            query += f"WHERE name = '{name}' AND variant = '{variant}';"
            is_update = True

        return (query, is_update)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: featureform/proto/serving.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1f\x66\x65\x61tureform/proto/serving.proto\x12\x19\x66\x65\x61tureform.serving.proto\"\x15\n\x05Model\x12\x0c\n\x04name\x18\x01 \x01(\t\"}\n\x13TrainingDataRequest\x12\x35\n\x02id\x18\x01 \x01(\x0b\x32).featureform.serving.proto.TrainingDataID\x12/\n\x05model\x18\x02 \x01(\x0b\x32 .featureform.serving.proto.Model\"/\n\x0eTrainingDataID\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"v\n\x0fTrainingDataRow\x12\x32\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32 .featureform.serving.proto.Value\x12/\n\x05label\x18\x02 \x01(\x0b\x32 .featureform.serving.proto.Value\"\xb3\x01\n\x13\x46\x65\x61tureServeRequest\x12\x36\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32$.featureform.serving.proto.FeatureID\x12\x33\n\x08\x65ntities\x18\x02 \x03(\x0b\x32!.featureform.serving.proto.Entity\x12/\n\x05model\x18\x03 \x01(\x0b\x32 .featureform.serving.proto.Model\">\n\nFeatureRow\x12\x30\n\x06values\x18\x01 \x03(\x0b\x32 .featureform.serving.proto.Value\"*\n\tFeatureID\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"%\n\x06\x45ntity\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x8a\x02\n\x05Value\x12\x13\n\tstr_value\x18\x01 \x01(\tH\x00\x12\x13\n\tint_value\x18\x02 \x01(\x05H\x00\x12\x15\n\x0b\x66loat_value\x18\x03 \x01(\x02H\x00\x12\x16\n\x0c\x64ouble_value\x18\x04 \x01(\x01H\x00\x12\x15\n\x0bint64_value\x18\x05 \x01(\x03H\x00\x12\x15\n\x0bint32_value\x18\x06 \x01(\x05H\x00\x12\x14\n\nbool_value\x18\x07 \x01(\x08H\x00\x12\x1c\n\x12on_demand_function\x18\x08 \x01(\x0cH\x00\x12=\n\x0evector32_value\x18\t \x01(\x0b\x32#.featureform.serving.proto.Vector32H\x00\x42\x07\n\x05value\")\n\x08SourceID\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"S\n\x11SourceDataRequest\x12/\n\x02id\x18\x01 \x01(\x0b\x32#.featureform.serving.proto.SourceID\x12\r\n\x05limit\x18\x02 \x01(\x03\"F\n\x13SourceColumnRequest\x12/\n\x02id\x18\x01 \x01(\x0b\x32#.featureform.serving.proto.SourceID\"?\n\rSourceDataRow\x12.\n\x04rows\x18\x01 \x03(\x0b\x32 .featureform.serving.proto.Value\"$\n\x11SourceDataColumns\x12\x0f\n\x07\x63olumns\x18\x01 \x03(\t\"S\n\x1aTrainingDataColumnsRequest\x12\x35\n\x02id\x18\x01 \x01(\x0b\x32).featureform.serving.proto.TrainingDataID\"2\n\x0fTrainingColumns\x12\x10\n\x08\x66\x65\x61tures\x18\x01 \x03(\t\x12\r\n\x05label\x18\x02 \x01(\t\"\x19\n\x08Vector32\x12\r\n\x05value\x18\x01 \x03(\x02\"\x82\x01\n\x0eNearestRequest\x12\x30\n\x02id\x18\x01 \x01(\x0b\x32$.featureform.serving.proto.FeatureID\x12\x33\n\x06vector\x18\x02 \x01(\x0b\x32#.featureform.serving.proto.Vector32\x12\t\n\x01k\x18\x03 \x01(\x05\"#\n\x0fNearestResponse\x12\x10\n\x08\x65ntities\x18\x01 \x03(\t2\x9d\x05\n\x07\x46\x65\x61ture\x12n\n\x0cTrainingData\x12..featureform.serving.proto.TrainingDataRequest\x1a*.featureform.serving.proto.TrainingDataRow\"\x00\x30\x01\x12z\n\x13TrainingDataColumns\x12\x35.featureform.serving.proto.TrainingDataColumnsRequest\x1a*.featureform.serving.proto.TrainingColumns\"\x00\x12g\n\x0c\x46\x65\x61tureServe\x12..featureform.serving.proto.FeatureServeRequest\x1a%.featureform.serving.proto.FeatureRow\"\x00\x12h\n\nSourceData\x12,.featureform.serving.proto.SourceDataRequest\x1a(.featureform.serving.proto.SourceDataRow\"\x00\x30\x01\x12o\n\rSourceColumns\x12..featureform.serving.proto.SourceColumnRequest\x1a,.featureform.serving.proto.SourceDataColumns\"\x00\x12\x62\n\x07Nearest\x12).featureform.serving.proto.NearestRequest\x1a*.featureform.serving.proto.NearestResponse\"\x00\x42\x1eZ\x1cgithub.com/featureform/protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'featureform.proto.serving_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\034github.com/featureform/proto'
  _globals['_MODEL']._serialized_start=62
  _globals['_MODEL']._serialized_end=83
  _globals['_TRAININGDATAREQUEST']._serialized_start=85
  _globals['_TRAININGDATAREQUEST']._serialized_end=210
  _globals['_TRAININGDATAID']._serialized_start=212
  _globals['_TRAININGDATAID']._serialized_end=259
  _globals['_TRAININGDATAROW']._serialized_start=261
  _globals['_TRAININGDATAROW']._serialized_end=379
  _globals['_FEATURESERVEREQUEST']._serialized_start=382
  _globals['_FEATURESERVEREQUEST']._serialized_end=561
  _globals['_FEATUREROW']._serialized_start=563
  _globals['_FEATUREROW']._serialized_end=625
  _globals['_FEATUREID']._serialized_start=627
  _globals['_FEATUREID']._serialized_end=669
  _globals['_ENTITY']._serialized_start=671
  _globals['_ENTITY']._serialized_end=708
  _globals['_VALUE']._serialized_start=711
  _globals['_VALUE']._serialized_end=977
  _globals['_SOURCEID']._serialized_start=979
  _globals['_SOURCEID']._serialized_end=1020
  _globals['_SOURCEDATAREQUEST']._serialized_start=1022
  _globals['_SOURCEDATAREQUEST']._serialized_end=1105
  _globals['_SOURCECOLUMNREQUEST']._serialized_start=1107
  _globals['_SOURCECOLUMNREQUEST']._serialized_end=1177
  _globals['_SOURCEDATAROW']._serialized_start=1179
  _globals['_SOURCEDATAROW']._serialized_end=1242
  _globals['_SOURCEDATACOLUMNS']._serialized_start=1244
  _globals['_SOURCEDATACOLUMNS']._serialized_end=1280
  _globals['_TRAININGDATACOLUMNSREQUEST']._serialized_start=1282
  _globals['_TRAININGDATACOLUMNSREQUEST']._serialized_end=1365
  _globals['_TRAININGCOLUMNS']._serialized_start=1367
  _globals['_TRAININGCOLUMNS']._serialized_end=1417
  _globals['_VECTOR32']._serialized_start=1419
  _globals['_VECTOR32']._serialized_end=1444
  _globals['_NEARESTREQUEST']._serialized_start=1447
  _globals['_NEARESTREQUEST']._serialized_end=1577
  _globals['_NEARESTRESPONSE']._serialized_start=1579
  _globals['_NEARESTRESPONSE']._serialized_end=1614
  _globals['_FEATURE']._serialized_start=1617
  _globals['_FEATURE']._serialized_end=2286
# @@protoc_insertion_point(module_scope)



import re

import pandas as pd


def get_sql_transformation_sources(query_string):
    # Use regex to parse query inputs in double curly braces {{ }}
    return [ts[2:-2].split(".") for ts in re.findall("(?={{).*?(?<=}})", query_string)]


def merge_feature_into_ts(feature_row, label_row, df, trainingset_df):
    if feature_row["source_timestamp"] != "":
        trainingset_df = pd.merge_asof(
            trainingset_df.sort_values(label_row["source_timestamp"]),
            df.sort_values(feature_row["source_timestamp"]),
            direction="backward",
            left_on=label_row["source_timestamp"],
            right_on=feature_row["source_timestamp"],
            left_by=label_row["source_entity"],
            right_by=feature_row["source_entity"],
        )
        if feature_row["source_timestamp"] != label_row["source_timestamp"]:
            trainingset_df.drop(columns=feature_row["source_timestamp"], inplace=True)
    else:
        df.drop_duplicates(
            subset=[feature_row["source_entity"]], keep="last", inplace=True
        )
        trainingset_df.reset_index(inplace=True)
        trainingset_df[label_row["source_entity"]] = trainingset_df[
            label_row["source_entity"]
        ].astype("string")
        df[label_row["source_entity"]] = df[label_row["source_entity"]].astype("string")
        trainingset_df = trainingset_df.join(
            df.set_index(label_row["source_entity"]),
            how="left",
            on=label_row["source_entity"],
            lsuffix="_left",
        )
        if "index" in trainingset_df.columns:
            trainingset_df.drop(columns="index", inplace=True)
    return trainingset_df


def list_to_combined_df(features_list, entity_id):
    all_feature_df = None
    try:
        for feature in features_list:
            if all_feature_df is None:
                all_feature_df = feature
            else:
                all_feature_df = all_feature_df.join(
                    feature.set_index(entity_id), on=entity_id, lsuffix="_left"
                )
        return all_feature_df
    except TypeError:
        print("Set is empty")


def feature_df_with_entity(source_path, entity_id, feature):
    name_variant = f"{feature['name']}.{feature['variant']}"
    df = pd.read_csv(str(source_path))
    check_missing_values(feature, df)
    if feature["source_timestamp"] != "":
        df = df[
            [
                feature["source_entity"],
                feature["source_value"],
                feature["source_timestamp"],
            ]
        ]
        df = df.sort_values(by=feature["source_timestamp"], ascending=True)
        df = df.drop(columns=feature["source_timestamp"])
    else:
        df = df[[feature["source_entity"], feature["source_value"]]]
    df.set_index(feature["source_entity"])
    df.rename(
        columns={
            feature["source_entity"]: entity_id,
            feature["source_value"]: name_variant,
        },
        inplace=True,
    )
    df.drop_duplicates(subset=[entity_id], keep="last", inplace=True)
    return df


def check_missing_values(resource, df):
    if resource["source_entity"] not in df.columns:
        raise KeyError(f"Entity column does not exist: {resource['source_entity']}")
    if resource["source_value"] not in df.columns:
        raise KeyError(f"Value column does not exist: {resource['source_value']}")
    if (
        resource["source_timestamp"] not in df.columns
        and resource["source_timestamp"] != ""
    ):
        raise KeyError(
            f"Timestamp column does not exist: {resource['source_timestamp']}"
        )


def label_df_from_csv(label, file_name):
    df = pd.read_csv(file_name)
    check_missing_values(label, df)
    if label["source_timestamp"] != "":
        df = df[
            [label["source_entity"], label["source_value"], label["source_timestamp"]]
        ]
        df[label["source_timestamp"]] = pd.to_datetime(df[label["source_timestamp"]])
        df.sort_values(by=label["source_timestamp"], inplace=True)
        df.drop_duplicates(
            subset=[label["source_entity"], label["source_timestamp"]],
            keep="last",
            inplace=True,
        )
    else:
        df = df[[label["source_entity"], label["source_value"]]]
    df.set_index(label["source_entity"], inplace=True)
    return df


def feature_df_from_csv(feature, filename):
    df = pd.read_csv(str(filename))
    check_missing_values(feature, df)
    if feature["source_timestamp"] != "":
        df = df[
            [
                feature["source_entity"],
                feature["source_value"],
                feature["source_timestamp"],
            ]
        ]
        df[feature["source_timestamp"]] = pd.to_datetime(
            df[feature["source_timestamp"]]
        )
        df = df.sort_values(by=feature["source_timestamp"], ascending=True)
    else:
        df = df[[feature["source_entity"], feature["source_value"]]]
    return df



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


import os
import base64
import json
import sys
import time
from abc import ABC, abstractmethod
from dataclasses import field
from typing import List, Tuple, Union, Optional

import dill
import grpc
from google.protobuf.duration_pb2 import Duration

from .enums import *
from .exceptions import *
from .sqlite_metadata import SQLiteMetadata
from .version import check_up_to_date

NameVariant = Tuple[str, str]

# Constants for Pyspark Versions
MAJOR_VERSION = "3"
MINOR_VERSIONS = ["7", "8", "9", "10", "11"]


@typechecked
def valid_name_variant(nvar: NameVariant) -> bool:
    return nvar[0] != "" and nvar[1] != ""


@typechecked
@dataclass
class Schedule:
    name: str
    variant: str
    resource_type: int
    schedule_string: str

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    def type(self) -> str:
        return "schedule"

    def _create(self, stub) -> None:
        serialized = pb.SetScheduleChangeRequest(
            resource=pb.ResourceId(
                pb.NameVariant(name=self.name, variant=self.variant),
                resource_type=self.resource_type,
            ),
            schedule=self.schedule_string,
        )
        stub.RequestScheduleChange(serialized)


@typechecked
@dataclass
class RedisConfig:
    host: str
    port: int
    password: str
    db: int

    def software(self) -> str:
        return "redis"

    def type(self) -> str:
        return "REDIS_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "Addr": f"{self.host}:{self.port}",
            "Password": self.password,
            "DB": self.db,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class PineconeConfig:
    project_id: str = ""
    environment: str = ""
    api_key: str = ""

    def software(self) -> str:
        return "pinecone"

    def type(self) -> str:
        return "PINECONE_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "ProjectID": self.project_id,
            "Environment": self.environment,
            "ApiKey": self.api_key,
        }
        return bytes(json.dumps(config), "utf-8")

    def deserialize(self, config):
        config = json.loads(config)
        self.project_id = config["ProjectID"]
        self.environment = config["Environment"]
        self.api_key = config["ApiKey"]
        return self


@typechecked
@dataclass
class WeaviateConfig:
    url: str = ""
    api_key: str = ""

    def software(self) -> str:
        return "weaviate"

    def type(self) -> str:
        return "WEAVIATE_ONLINE"

    def serialize(self) -> bytes:
        if self.url == "":
            raise Exception("URL cannot be empty")
        config = {
            "URL": self.url,
            "ApiKey": self.api_key,
        }
        return bytes(json.dumps(config), "utf-8")

    def deserialize(self, config):
        config = json.loads(config)
        self.url = config["URL"]
        self.api_key = config["ApiKey"]
        return self


@typechecked
@dataclass
class AWSCredentials:
    def __init__(
        self,
        access_key: str,
        secret_key: str,
    ):
        """

        Credentials for an AWS.

        **Example**
        ```
        aws_credentials = ff.AWSCredentials(
            access_key="<AWS_ACCESS_KEY>",
            secret_key="<AWS_SECRET_KEY>"
        )
        ```

        Args:
            access_key (str): AWS Access Key.
            secret_key (str): AWS Secret Key.
        """
        if access_key == "":
            raise Exception("'AWSCredentials' access_key cannot be empty")

        if secret_key == "":
            raise Exception("'AWSCredentials' secret_key cannot be empty")

        self.access_key = access_key
        self.secret_key = secret_key

    def type(self):
        return "AWS_CREDENTIALS"

    def config(self):
        return {
            "AWSAccessKeyId": self.access_key,
            "AWSSecretKey": self.secret_key,
        }


@typechecked
@dataclass
class GCPCredentials:
    def __init__(
        self,
        project_id: str,
        credentials_path: str,
    ):
        """

        Credentials for an GCP.

        **Example**
        ```
        gcp_credentials = ff.GCPCredentials(
            project_id="<project_id>",
            credentials_path="<path_to_credentials>"
        )
        ```

        Args:
            project_id (str): The project id.
            credentials_path (str): The path to the credentials file.
        """
        if project_id == "":
            raise Exception("'GCPCredentials' project_id cannot be empty")

        if credentials_path == "":
            raise Exception("'GCPCredentials' credentials_path cannot be empty")

        if not os.path.isfile(credentials_path):
            raise Exception(
                f"'GCPCredentials' credentials_path '{credentials_path}' file not found"
            )

        self.project_id = project_id
        with open(credentials_path) as f:
            self.credentials = json.load(f)

    def type(self):
        return "GCPCredentials"

    def config(self):
        return {
            "ProjectId": self.project_id,
            "JSON": self.credentials,
        }

    def to_json(self):
        return self.credentials


@typechecked
@dataclass
class GCSFileStoreConfig:
    credentials: GCPCredentials
    bucket_name: str
    bucket_path: str = ""

    def software(self) -> str:
        return "gcs"

    def type(self) -> str:
        return "GCS"

    def serialize(self) -> bytes:
        config = {
            "BucketName": self.bucket_name,
            "BucketPath": self.bucket_path,
            "Credentials": self.credentials.config(),
        }
        return bytes(json.dumps(config), "utf-8")

    def config(self):
        return {
            "BucketName": self.bucket_name,
            "BucketPath": self.bucket_path,
            "Credentials": self.credentials.config(),
        }

    def store_type(self):
        return self.type()


@typechecked
@dataclass
class AzureFileStoreConfig:
    account_name: str
    account_key: str
    container_name: str
    root_path: str

    def software(self) -> str:
        return "azure"

    def type(self) -> str:
        return "AZURE"

    def serialize(self) -> bytes:
        config = {
            "AccountName": self.account_name,
            "AccountKey": self.account_key,
            "ContainerName": self.container_name,
            "Path": self.root_path,
        }
        return bytes(json.dumps(config), "utf-8")

    def config(self):
        return {
            "AccountName": self.account_name,
            "AccountKey": self.account_key,
            "ContainerName": self.container_name,
            "Path": self.root_path,
        }

    def store_type(self):
        return self.type()


@typechecked
@dataclass
class S3StoreConfig:
    def __init__(
        self,
        bucket_path: str,
        bucket_region: str,
        credentials: AWSCredentials,
        path: str = "",
    ):
        bucket_path_ends_with_slash = len(bucket_path) != 0 and bucket_path[-1] == "/"

        if bucket_path_ends_with_slash:
            raise Exception("The 'bucket_path' cannot end with '/'.")

        self.bucket_path = bucket_path
        self.bucket_region = bucket_region
        self.credentials = credentials
        self.path = path

    def software(self) -> str:
        return "S3"

    def type(self) -> str:
        return "S3"

    def serialize(self) -> bytes:
        config = self.config()
        return bytes(json.dumps(config), "utf-8")

    def config(self):
        return {
            "Credentials": self.credentials.config(),
            "BucketRegion": self.bucket_region,
            "BucketPath": self.bucket_path,
            "Path": self.path,
        }

    def store_type(self):
        return self.type()


@typechecked
@dataclass
class HDFSConfig:
    def __init__(self, host: str, port: str, path: str, username: str):
        bucket_path_ends_with_slash = len(path) != 0 and path[-1] == "/"

        if bucket_path_ends_with_slash:
            raise Exception("The 'bucket_path' cannot end with '/'.")

        self.path = path
        self.host = host
        self.port = port
        self.username = username

    def software(self) -> str:
        return "HDFS"

    def type(self) -> str:
        return "HDFS"

    def serialize(self) -> bytes:
        config = {
            "Host": self.host,
            "Port": self.port,
            "Path": self.path,
            "Username": self.username,
        }
        return bytes(json.dumps(config), "utf-8")

    def config(self):
        return {
            "Host": self.host,
            "Port": self.port,
            "Path": self.path,
            "Username": self.username,
        }

    def store_type(self):
        return self.type()


@typechecked
@dataclass
class OnlineBlobConfig:
    store_type: str
    store_config: dict

    def software(self) -> str:
        return self.store_type

    def type(self) -> str:
        return "BLOB_ONLINE"

    def config(self):
        return self.store_config

    def serialize(self) -> bytes:
        config = {
            "Type": self.store_type,
            "Config": self.store_config,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class FirestoreConfig:
    collection: str
    project_id: str
    credentials: GCPCredentials

    def software(self) -> str:
        return "firestore"

    def type(self) -> str:
        return "FIRESTORE_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "Collection": self.collection,
            "ProjectID": self.project_id,
            "Credentials": self.credentials.to_json(),
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class CassandraConfig:
    keyspace: str
    host: str
    port: int
    username: str
    password: str
    consistency: str
    replication: int

    def software(self) -> str:
        return "cassandra"

    def type(self) -> str:
        return "CASSANDRA_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "Keyspace": self.keyspace,
            "Addr": f"{self.host}:{self.port}",
            "Username": self.username,
            "Password": self.password,
            "Consistency": self.consistency,
            "Replication": self.replication,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class DynamodbConfig:
    region: str
    access_key: str
    secret_key: str

    def software(self) -> str:
        return "dynamodb"

    def type(self) -> str:
        return "DYNAMODB_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "Region": self.region,
            "AccessKey": self.access_key,
            "SecretKey": self.secret_key,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class MongoDBConfig:
    username: str
    password: str
    host: str
    port: str
    database: str
    throughput: int

    def software(self) -> str:
        return "mongodb"

    def type(self) -> str:
        return "MONGODB_ONLINE"

    def serialize(self) -> bytes:
        config = {
            "Username": self.username,
            "Password": self.password,
            "Host": self.host,
            "Port": self.port,
            "Database": self.database,
            "Throughput": self.throughput,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class SnowflakeConfig:
    username: str
    password: str
    schema: str
    account: str = ""
    organization: str = ""
    database: str = ""
    account_locator: str = ""
    warehouse: str = ""
    role: str = ""

    def __post_init__(self):
        if self.__has_legacy_credentials() and self.__has_current_credentials():
            raise ValueError(
                "Cannot create configure Snowflake with both current and legacy credentials"
            )

        if not self.__has_legacy_credentials() and not self.__has_current_credentials():
            raise ValueError("Cannot create configure Snowflake without credentials")

    def __has_legacy_credentials(self) -> bool:
        return self.account_locator != ""

    def __has_current_credentials(self) -> bool:
        if (self.account != "" and self.organization == "") or (
            self.account == "" and self.organization != ""
        ):
            raise ValueError("Both Snowflake organization and account must be included")
        elif self.account != "" and self.organization != "":
            return True
        else:
            return False

    def software(self) -> str:
        return "Snowflake"

    def type(self) -> str:
        return "SNOWFLAKE_OFFLINE"

    def serialize(self) -> bytes:
        config = {
            "Username": self.username,
            "Password": self.password,
            "Organization": self.organization,
            "AccountLocator": self.account_locator,
            "Account": self.account,
            "Schema": self.schema,
            "Database": self.database,
            "Warehouse": self.warehouse,
            "Role": self.role,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class PostgresConfig:
    host: str
    port: str
    database: str
    user: str
    password: str
    sslmode: str

    def software(self) -> str:
        return "postgres"

    def type(self) -> str:
        return "POSTGRES_OFFLINE"

    def serialize(self) -> bytes:
        config = {
            "Host": self.host,
            "Port": self.port,
            "Username": self.user,
            "Password": self.password,
            "Database": self.database,
            "SSLMode": self.sslmode,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class RedshiftConfig:
    host: str
    port: int
    database: str
    user: str
    password: str

    def software(self) -> str:
        return "redshift"

    def type(self) -> str:
        return "REDSHIFT_OFFLINE"

    def serialize(self) -> bytes:
        config = {
            "Host": self.host,
            "Port": str(self.port),
            "Username": self.user,
            "Password": self.password,
            "Database": self.database,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class BigQueryConfig:
    project_id: str
    dataset_id: str
    credentials: GCPCredentials

    def software(self) -> str:
        return "bigquery"

    def type(self) -> str:
        return "BIGQUERY_OFFLINE"

    def serialize(self) -> bytes:
        config = {
            "ProjectID": self.project_id,
            "DatasetID": self.dataset_id,
            "Credentials": self.credentials.to_json(),
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class SparkConfig:
    executor_type: str
    executor_config: dict
    store_type: str
    store_config: dict

    def software(self) -> str:
        return "spark"

    def type(self) -> str:
        return "SPARK_OFFLINE"

    def serialize(self) -> bytes:
        config = {
            "ExecutorType": self.executor_type,
            "StoreType": self.store_type,
            "ExecutorConfig": self.executor_config,
            "StoreConfig": self.store_config,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class K8sResourceSpecs:
    cpu_request: str = ""
    cpu_limit: str = ""
    memory_request: str = ""
    memory_limit: str = ""


@typechecked
@dataclass
class K8sArgs:
    docker_image: str
    specs: Union[K8sResourceSpecs, None] = None

    def apply(self, transformation: pb.Transformation):
        transformation.kubernetes_args.docker_image = self.docker_image
        if self.specs is not None:
            transformation.kubernetes_args.specs.cpu_request = self.specs.cpu_request
            transformation.kubernetes_args.specs.cpu_limit = self.specs.cpu_limit
            transformation.kubernetes_args.specs.memory_request = (
                self.specs.memory_request
            )
            transformation.kubernetes_args.specs.memory_limit = self.specs.memory_limit
        return transformation


@typechecked
@dataclass
class K8sConfig:
    store_type: str
    store_config: dict
    docker_image: str = ""

    def software(self) -> str:
        return "k8s"

    def type(self) -> str:
        return "K8S_OFFLINE"

    def serialize(self) -> bytes:
        config = {
            "ExecutorType": "K8S",
            "ExecutorConfig": {"docker_image": self.docker_image},
            "StoreType": self.store_type,
            "StoreConfig": self.store_config,
        }
        return bytes(json.dumps(config), "utf-8")


@typechecked
@dataclass
class EmptyConfig:
    def software(self) -> str:
        return ""

    def type(self) -> str:
        return ""

    def serialize(self) -> bytes:
        return bytes("", "utf-8")

    def deserialize(self, config):
        return self


@typechecked
@dataclass
class LocalConfig:
    def software(self) -> str:
        return "localmode"

    def type(self) -> str:
        return "LOCAL_ONLINE"

    def serialize(self) -> bytes:
        return bytes(json.dumps({}), "utf-8")


Config = Union[
    RedisConfig,
    PineconeConfig,
    SnowflakeConfig,
    PostgresConfig,
    RedshiftConfig,
    PineconeConfig,
    LocalConfig,
    BigQueryConfig,
    FirestoreConfig,
    SparkConfig,
    OnlineBlobConfig,
    AzureFileStoreConfig,
    S3StoreConfig,
    K8sConfig,
    MongoDBConfig,
    GCSFileStoreConfig,
    EmptyConfig,
    HDFSConfig,
    WeaviateConfig,
    DynamodbConfig,
    CassandraConfig,
]


@typechecked
@dataclass
class Properties:
    properties: dict

    def __post_init__(self):
        self.serialized = pb.Properties()
        for key, val in self.properties.items():
            self.serialized.property[key].string_value = val


@typechecked
@dataclass
class Provider:
    name: str
    description: str
    team: str
    config: Config
    function: str
    status: str = "NO_STATUS"
    tags: list = field(default_factory=list)
    properties: dict = field(default_factory=dict)
    error: Optional[str] = None

    def __post_init__(self):
        self.software = self.config.software() if self.config is not None else None

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def type() -> str:
        return "provider"

    def get(self, stub) -> "Provider":
        name = pb.Name(name=self.name)
        provider = next(stub.GetProviders(iter([name])))

        return Provider(
            name=provider.name,
            description=provider.description,
            function=provider.type,
            team=provider.team,
            config=EmptyConfig(),  # TODO add deserializer to configs
            tags=list(provider.tags.tag),
            properties={k: v for k, v in provider.properties.property.items()},
            status=provider.status.Status._enum_type.values[
                provider.status.status
            ].name,
            error=provider.status.error_message,
        )

    def _create(self, stub) -> None:
        serialized = pb.Provider(
            name=self.name,
            description=self.description,
            type=self.config.type(),
            software=self.config.software(),
            team=self.team,
            serialized_config=self.config.serialize(),
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
        )
        stub.CreateProvider(serialized)

    def _create_local(self, db) -> None:
        db.insert(
            "providers",
            self.name,
            "Provider",
            self.description,
            self.config.type(),
            self.config.software(),
            self.team,
            "sources",
            "ready",
            str(self.config.serialize(), "utf-8"),
        )
        if len(self.tags):
            db.upsert("tags", self.name, "", "providers", json.dumps(self.tags))
        if len(self.properties):
            db.upsert(
                "properties", self.name, "", "providers", json.dumps(self.properties)
            )

    def __eq__(self, other):
        for attribute in vars(self):
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True

    def to_dictionary(self):
        return {
            "name": self.name,
            "description": self.description,
            "team": self.team,
            "config": "todox",
            "function": "todox",
            "status": self.status,
            "tags": self.tags,
            "properties": self.properties,
            "error": self.error,
        }


@typechecked
@dataclass
class User:
    name: str
    status: str = ""
    tags: list = field(default_factory=list)
    properties: dict = field(default_factory=dict)

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    def type(self) -> str:
        return "user"

    def _create(self, stub) -> None:
        serialized = pb.User(
            name=self.name,
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
        )
        stub.CreateUser(serialized)

    def _create_local(self, db) -> None:
        db.insert("users", self.name, "User", "ready")
        if len(self.tags):
            db.upsert("tags", self.name, "", "users", json.dumps(self.tags))
        if len(self.properties):
            db.upsert("properties", self.name, "", "users", json.dumps(self.properties))

    def __eq__(self, other):
        for attribute in vars(self):
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True

    def to_dictionary(self):
        return {
            "name": self.name,
            "status": self.status,
            "tags": self.tags,
            "properties": self.properties,
        }


@typechecked
@dataclass
class SQLTable:
    name: str


@typechecked
@dataclass
class Directory:
    path: str


Location = Union[SQLTable, Directory]


@typechecked
@dataclass
class PrimaryData:
    location: Location

    def kwargs(self):
        return {
            "primaryData": pb.PrimaryData(
                table=pb.PrimarySQLTable(
                    name=self.location.name,
                ),
            ),
        }

    def name(self):
        return self.location.name

    def path(self):
        return self.location.path


class Transformation:
    pass


@typechecked
@dataclass
class SQLTransformation(Transformation):
    query: str
    args: K8sArgs = None

    def type(self):
        return SourceType.SQL_TRANSFORMATION.value

    def kwargs(self):
        transformation = pb.Transformation(
            SQLTransformation=pb.SQLTransformation(
                query=self.query,
            )
        )

        if self.args is not None:
            transformation = self.args.apply(transformation)

        return {"transformation": transformation}


@typechecked
@dataclass
class DFTransformation(Transformation):
    query: bytes
    inputs: list
    args: K8sArgs = None
    source_text: str = ""

    def type(self):
        return SourceType.DF_TRANSFORMATION.value

    def kwargs(self):
        transformation = pb.Transformation(
            DFTransformation=pb.DFTransformation(
                query=self.query,
                inputs=[pb.NameVariant(name=v[0], variant=v[1]) for v in self.inputs],
                source_text=self.source_text,
            )
        )

        if self.args is not None:
            transformation = self.args.apply(transformation)

        return {"transformation": transformation}


SourceDefinition = Union[PrimaryData, Transformation, str]


@typechecked
@dataclass
class Source:
    name: str
    default_variant: str
    variants: List[str]

    def to_dictionary(self):
        return {
            "name": self.name,
            "default_variant": self.default_variant,
            "variants": self.variants,
        }


@typechecked
@dataclass
class SourceVariant:
    name: str
    definition: SourceDefinition
    owner: str
    provider: str
    description: str
    tags: list
    properties: dict
    variant: str
    created: str = None
    status: str = "ready"  # this is no status by default but it always stores ready
    schedule: str = ""
    schedule_obj: Schedule = None
    is_transformation = SourceType.PRIMARY_SOURCE.value
    source_text: str = ""
    source_type: str = ""
    transformation: str = ""
    inputs = ([],)
    error: Optional[str] = None

    def update_schedule(self, schedule) -> None:
        self.schedule_obj = Schedule(
            name=self.name,
            variant=self.variant,
            resource_type=7,
            schedule_string=schedule,
        )
        self.schedule = schedule

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def type() -> str:
        return "source"

    def get(self, stub):
        name_variant = pb.NameVariant(name=self.name, variant=self.variant)
        source = next(stub.GetSourceVariants(iter([name_variant])))
        definition = self._get_source_definition(source)

        return SourceVariant(
            created=None,
            name=source.name,
            definition=definition,
            owner=source.owner,
            provider=source.provider,
            description=source.description,
            variant=source.variant,
            tags=list(source.tags.tag),
            properties={k: v for k, v in source.properties.property.items()},
            status=source.status.Status._enum_type.values[source.status.status].name,
            error=source.status.error_message,
        )

    def _get_source_definition(self, source):
        if source.primaryData.table.name:
            return PrimaryData(SQLTable(source.primaryData.table.name))
        elif source.transformation:
            return self._get_transformation_definition(source)
        else:
            raise Exception(f"Invalid source type {source}")

    def _get_transformation_definition(self, source):
        if source.transformation.DFTransformation.query != bytes():
            transformation = source.transformation.DFTransformation
            return DFTransformation(
                query=transformation.query,
                inputs=[(input.name, input.variant) for input in transformation.inputs],
                source_text=transformation.source_text,
            )
        elif source.transformation.SQLTransformation.query != "":
            return SQLTransformation(source.transformation.SQLTransformation.query)
        else:
            raise Exception(f"Invalid transformation type {source}")

    def _create(self, stub) -> None:
        defArgs = self.definition.kwargs()
        serialized = pb.SourceVariant(
            created=None,
            name=self.name,
            variant=self.variant,
            owner=self.owner,
            description=self.description,
            schedule=self.schedule,
            provider=self.provider,
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
            **defArgs,
        )
        stub.CreateSourceVariant(serialized)

    def _create_local(self, db) -> None:
        should_insert_text = False
        source_text = ""
        self.source_type = "Source"
        if type(self.definition) == DFTransformation:
            should_insert_text = True
            self.is_transformation = SourceType.DF_TRANSFORMATION.value
            source_text = self.definition.source_text
            self.inputs = self.definition.inputs
            self.definition = self.definition.query
            self.source_text = source_text
            self.source_type = "Dataframe Transformation"
        elif type(self.definition) == SQLTransformation:
            self.is_transformation = SourceType.SQL_TRANSFORMATION.value
            self.definition = self.definition.query
        elif type(self.definition) == PrimaryData:
            if isinstance(self.definition.location, Directory):
                self.definition = self.definition.path()
                self.is_transformation = SourceType.DIRECTORY.value
            elif isinstance(self.definition.location, SQLTable):
                self.definition = self.definition.name()
                self.is_transformation = SourceType.PRIMARY_SOURCE.value
            else:
                raise ValueError(
                    f"Invalid Primary Data Type {self.definition.location}"
                )
        db.insert_source(
            "source_variant",
            str(time.time()),
            self.description,
            self.name,
            self.source_type,
            self.owner,
            self.provider,
            self.variant,
            "ready",
            self.is_transformation,
            json.dumps(self.inputs),
            self.definition,
        )

        if should_insert_text:
            db.insert_source_variant_text(
                str(time.time()), self.name, self.variant, source_text
            )

        if len(self.tags):
            db.upsert(
                "tags", self.name, self.variant, "source_variant", json.dumps(self.tags)
            )
        if len(self.properties):
            db.upsert(
                "properties",
                self.name,
                self.variant,
                "source_variant",
                json.dumps(self.properties),
            )
        self._create_source_resource(db)

    def _create_source_resource(self, db) -> None:
        db.insert("sources", "Source", self.variant, self.name)

    def get_status(self):
        return ResourceStatus(self.status)

    def is_ready(self):
        return self.status == ResourceStatus.READY.value

    def __eq__(self, other):
        for attribute in vars(self):
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True


@typechecked
@dataclass
class Entity:
    name: str
    description: str
    status: str = "NO_STATUS"
    tags: list = field(default_factory=list)
    properties: dict = field(default_factory=dict)

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def type() -> str:
        return "entity"

    def _create(self, stub) -> None:
        serialized = pb.Entity(
            name=self.name,
            description=self.description,
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
        )
        stub.CreateEntity(serialized)

    def _create_local(self, db) -> None:
        db.insert("entities", self.name, "Entity", self.description, "ready")
        if len(self.tags):
            db.upsert("tags", self.name, "", "entities", json.dumps(self.tags))
        if len(self.properties):
            db.upsert(
                "properties", self.name, "", "entities", json.dumps(self.properties)
            )

    def __eq__(self, other):
        for attribute in vars(self):
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True

    def to_dictionary(self):
        return {
            "name": self.name,
            "description": self.description,
            "status": self.status,
            "tags": self.tags,
            "properties": self.properties,
        }


@typechecked
@dataclass
class ResourceColumnMapping:
    entity: str
    value: str
    timestamp: str

    def proto(self) -> pb.Columns:
        return pb.Columns(
            entity=self.entity,
            value=self.value,
            ts=self.timestamp,
        )


ResourceLocation = ResourceColumnMapping


@typechecked
@dataclass
class Feature:
    name: str
    default_variant: str
    variants: List[str]

    def to_dictionary(self):
        return {
            "name": self.name,
            "default_variant": self.default_variant,
            "variants": self.variants,
        }


@typechecked
@dataclass
class FeatureVariant:
    name: str
    source: NameVariant
    value_type: str
    entity: str
    owner: str
    provider: str
    location: ResourceLocation
    description: str
    variant: str
    created: str = None
    is_embedding: bool = False
    dims: int = 0
    tags: list = None
    properties: dict = None
    schedule: str = ""
    schedule_obj: Schedule = None
    status: str = "NO_STATUS"
    error: Optional[str] = None

    def __post_init__(self):
        col_types = [member.value for member in ScalarType]
        if self.value_type not in col_types:
            raise ValueError(
                f"Invalid feature type ({self.value_type}) must be one of: {col_types}"
            )

    def update_schedule(self, schedule) -> None:
        self.schedule_obj = Schedule(
            name=self.name,
            variant=self.variant,
            resource_type=4,
            schedule_string=schedule,
        )
        self.schedule = schedule

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def type() -> str:
        return "feature"

    def get(self, stub) -> "FeatureVariant":
        name_variant = pb.NameVariant(name=self.name, variant=self.variant)
        feature = next(stub.GetFeatureVariants(iter([name_variant])))

        return FeatureVariant(
            created=None,
            name=feature.name,
            variant=feature.variant,
            source=(feature.source.name, feature.source.variant),
            value_type=feature.type,
            is_embedding=feature.is_embedding,
            dims=feature.dimension,
            entity=feature.entity,
            owner=feature.owner,
            provider=feature.provider,
            location=ResourceColumnMapping("", "", ""),
            description=feature.description,
            tags=list(feature.tags.tag),
            properties={k: v for k, v in feature.properties.property.items()},
            status=feature.status.Status._enum_type.values[feature.status.status].name,
            error=feature.status.error_message,
        )

    def _create(self, stub) -> None:
        serialized = pb.FeatureVariant(
            name=self.name,
            variant=self.variant,
            source=pb.NameVariant(
                name=self.source[0],
                variant=self.source[1],
            ),
            type=self.value_type,
            is_embedding=self.is_embedding,
            dimension=self.dims,
            entity=self.entity,
            owner=self.owner,
            description=self.description,
            schedule=self.schedule,
            provider=self.provider,
            columns=self.location.proto(),
            mode=ComputationMode.PRECOMPUTED.proto(),
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
        )
        stub.CreateFeatureVariant(serialized)

    def _create_local(self, db) -> None:
        db.insert(
            "feature_variant",
            str(time.time()),
            self.description,
            self.entity,
            self.name,
            self.owner,
            self.provider,
            self.value_type,
            self.variant,
            "ready",
            self.location.entity,
            self.location.timestamp,
            self.location.value,
            self.source[0],
            self.source[1],
            self.is_embedding,
            self.dims,
        )
        if len(self.tags):
            db.upsert(
                "tags",
                self.name,
                self.variant,
                "feature_variant",
                json.dumps(self.tags),
            )
        if len(self.properties):
            db.upsert(
                "properties",
                self.name,
                self.variant,
                "feature_variant",
                json.dumps(self.properties),
            )

        self._write_feature_variant_and_mode(db)

    def _write_feature_variant_and_mode(self, db) -> None:
        db.insert(
            "features",
            self.name,
            self.variant,
            self.value_type,
        )
        is_on_demand = 0
        db.insert(
            "feature_computation_mode",
            self.name,
            self.variant,
            ComputationMode.PRECOMPUTED.value,
            is_on_demand,
        )

    def get_status(self):
        return ResourceStatus(self.status)

    def is_ready(self):
        return self.status == ResourceStatus.READY.value

    def __eq__(self, other):
        for attribute in vars(self):
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True


class BaseStream(ABC):
    def __init__(
        self,
        name: str,
        value_type: str,
        entity: str,
        owner: str,
        offline_provider: str,
        description: str,
        variant: str,
        tags: Union[list, None] = None,
        properties: Union[dict, None] = None,
        status: str = "NO_STATUS",
        error: Optional[str] = None,
    ):
        self.name = name
        self.value_type = value_type
        self.entity = entity
        self.owner = owner
        self.offline_provider = offline_provider
        self.description = description
        self.variant = variant
        self.tags = [] if tags is None else tags
        self.properties = {} if properties is None else properties
        self.status = status
        self.error = error

    def __post_init__(self):
        col_types = [member.value for member in ScalarType]
        if self.value_type not in col_types:
            raise ValueError(
                f"Invalid feature type ({self.value_type}) must be one of: {col_types}"
            )

    @staticmethod
    @abstractmethod
    def type() -> str:
        raise NotImplementedError

    @abstractmethod
    def get(self, stub) -> "Stream":
        raise NotImplementedError

    @abstractmethod
    def _create(self, stub) -> None:
        raise NotImplementedError

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    def _create_local(self, db) -> None:
        pass

    def get_status(self):
        return ResourceStatus(self.status)

    def is_ready(self):
        return self.status == ResourceStatus.READY.value

    def __eq__(self, other):
        for attribute in vars(self):
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True


@typechecked
class StreamFeature(BaseStream):
    inference_store: str = ""

    def __init__(self, inference_store: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inference_store = inference_store

    @staticmethod
    def type() -> str:
        return "feature"

    def get(self, stub) -> "StreamFeature":
        name_variant = pb.NameVariant(name=self.name, variant=self.variant)
        stream_feature = next(stub.GetFeatureVariants(iter([name_variant])))

        return StreamFeature(
            name=stream_feature.name,
            variant=stream_feature.variant,
            value_type=stream_feature.type,
            entity=stream_feature.entity,
            owner=stream_feature.owner,
            inference_store=stream_feature.provider,
            offline_provider=stream_feature.stream.offline_provider,
            description=stream_feature.description,
            tags=list(stream_feature.tags.tag),
            properties={k: v for k, v in stream_feature.properties.property.items()},
            status=stream_feature.status.Status._enum_type.values[
                stream_feature.status.status
            ].name,
            error=stream_feature.status.error_message,
        )

    def _create(self, stub) -> None:
        serialized = pb.FeatureVariant(
            name=self.name,
            variant=self.variant,
            type=self.value_type,
            entity=self.entity,
            owner=self.owner,
            description=self.description,
            provider=self.inference_store,
            stream=pb.Stream(
                offline_provider=self.offline_provider,
            ),
            mode=ComputationMode.STREAMING.proto(),
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
        )
        stub.CreateFeatureVariant(serialized)


@typechecked
class StreamLabel(BaseStream):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def type() -> str:
        return "label"

    def get(self, stub) -> "StreamLabel":
        name_variant = pb.NameVariant(name=self.name, variant=self.variant)
        stream_label = next(stub.GetLabelVariants(iter([name_variant])))

        return StreamLabel(
            name=stream_label.name,
            variant=stream_label.variant,
            value_type=stream_label.type,
            entity=stream_label.entity,
            owner=stream_label.owner,
            offline_provider=stream_label.stream.offline_provider,
            description=stream_label.description,
            tags=list(stream_label.tags.tag),
            properties={k: v for k, v in stream_label.properties.property.items()},
            status=stream_label.status.Status._enum_type.values[
                stream_label.status.status
            ].name,
            error=stream_label.status.error_message,
        )

    def _create(self, stub) -> None:
        serialized = pb.LabelVariant(
            name=self.name,
            variant=self.variant,
            type=self.value_type,
            entity=self.entity,
            owner=self.owner,
            description=self.description,
            provider=self.offline_provider,
            stream=pb.Stream(
                offline_provider=self.offline_provider,
            ),
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
        )
        stub.CreateLabelVariant(serialized)


@typechecked
@dataclass
class OnDemandFeatureVariant:
    owner: str
    variant: str
    tags: List[str] = field(default_factory=list)
    properties: dict = field(default_factory=dict)
    name: str = ""
    description: str = ""
    status: str = "READY"
    error: Optional[str] = None

    def __call__(self, fn):
        if self.description == "" and fn.__doc__ is not None:
            self.description = fn.__doc__
        if self.name == "":
            self.name = fn.__name__

        self.query = dill.dumps(fn.__code__)
        fn.name_variant = self.name_variant
        fn.query = self.query
        return fn

    def name_variant(self):
        return (self.name, self.variant)

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def type() -> str:
        return "ondemand_feature"

    def _create(self, stub) -> None:
        serialized = pb.FeatureVariant(
            name=self.name,
            variant=self.variant,
            owner=self.owner,
            description=self.description,
            function=pb.PythonFunction(query=self.query),
            mode=ComputationMode.CLIENT_COMPUTED.proto(),
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
            status=pb.ResourceStatus(status=pb.ResourceStatus.READY),
        )
        stub.CreateFeatureVariant(serialized)

    def _create_local(self, db) -> None:
        decode_query = base64.b64encode(self.query).decode("ascii")

        db.insert(
            "ondemand_feature_variant",
            str(time.time()),
            self.description,
            self.name,
            self.owner,
            self.variant,
            "ready",
            decode_query,
        )
        if self.tags and len(self.tags):
            db.upsert(
                "tags",
                self.name,
                self.variant,
                "feature_variant",
                json.dumps(self.tags),
            )
        if len(self.properties):
            db.upsert(
                "properties",
                self.name,
                self.variant,
                "feature_variant",
                json.dumps(self.properties),
            )

        self._write_feature_variant_and_mode(db)

    def _write_feature_variant_and_mode(self, db) -> None:
        db.insert(
            "features",
            self.name,
            self.variant,
            "tbd",
        )
        is_on_demand = 1

        db.insert(
            "feature_computation_mode",
            self.name,
            self.variant,
            ComputationMode.CLIENT_COMPUTED.value,
            is_on_demand,
        )

    def get(self, stub) -> "OnDemandFeatureVariant":
        name_variant = pb.NameVariant(name=self.name, variant=self.variant)
        ondemand_feature = next(stub.GetFeatureVariants(iter([name_variant])))

        return OnDemandFeatureVariant(
            name=ondemand_feature.name,
            variant=ondemand_feature.variant,
            owner=ondemand_feature.owner,
            description=ondemand_feature.description,
            tags=list(ondemand_feature.tags.tag),
            properties={k: v for k, v in ondemand_feature.properties.property.items()},
            status=ondemand_feature.status.Status._enum_type.values[
                ondemand_feature.status.status
            ].name,
            error=ondemand_feature.status.error_message,
        )

    def get_status(self):
        return ResourceStatus(self.status)

    def is_ready(self):
        return self.status == ResourceStatus.READY.value

    def __eq__(self, other):
        for attribute in vars(self):
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True


@typechecked
@dataclass
class Label:
    name: str
    default_variant: str
    variants: List[str]

    def to_dictionary(self):
        return {
            "name": self.name,
            "default_variant": self.default_variant,
            "variants": self.variants,
        }


@typechecked
@dataclass
class LabelVariant:
    name: str
    source: NameVariant
    value_type: str
    entity: str
    owner: str
    provider: str
    description: str
    tags: list
    properties: dict
    location: ResourceLocation
    variant: str
    created: str = None
    status: str = "NO_STATUS"
    error: Optional[str] = None

    def __post_init__(self):
        col_types = [member.value for member in ScalarType]
        if self.value_type not in col_types:
            raise ValueError(
                f"Invalid label type ({self.value_type}) must be one of: {col_types}"
            )

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def type() -> str:
        return "label"

    def get(self, stub) -> "LabelVariant":
        name_variant = pb.NameVariant(name=self.name, variant=self.variant)
        label = next(stub.GetLabelVariants(iter([name_variant])))

        return LabelVariant(
            name=label.name,
            variant=label.variant,
            source=(label.source.name, label.source.variant),
            value_type=label.type,
            entity=label.entity,
            owner=label.owner,
            provider=label.provider,
            location=ResourceColumnMapping("", "", ""),
            description=label.description,
            tags=list(label.tags.tag),
            properties={k: v for k, v in label.properties.property.items()},
            status=label.status.Status._enum_type.values[label.status.status].name,
            error=label.status.error_message,
        )

    def _create(self, stub) -> None:
        serialized = pb.LabelVariant(
            name=self.name,
            variant=self.variant,
            source=pb.NameVariant(
                name=self.source[0],
                variant=self.source[1],
            ),
            provider=self.provider,
            type=self.value_type,
            entity=self.entity,
            owner=self.owner,
            description=self.description,
            columns=self.location.proto(),
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
        )
        stub.CreateLabelVariant(serialized)

    def _create_local(self, db) -> None:
        db.insert(
            "label_variant",
            str(time.time()),
            self.description,
            self.entity,
            self.name,
            self.owner,
            self.provider,
            self.value_type,
            self.variant,
            self.location.entity,
            self.location.timestamp,
            self.location.value,
            "ready",
            self.source[0],
            self.source[1],
        )
        if len(self.tags):
            db.upsert(
                "tags", self.name, self.variant, "label_variant", json.dumps(self.tags)
            )
        if len(self.properties):
            db.upsert(
                "properties",
                self.name,
                self.variant,
                "label_variant",
                json.dumps(self.properties),
            )
        self._create_label_resource(db)

    def _create_label_resource(self, db) -> None:
        db.insert("labels", self.value_type, self.variant, self.name)

    def get_status(self):
        return ResourceStatus(self.status)

    def is_ready(self):
        return self.status == ResourceStatus.READY.value

    def __eq__(self, other):
        for attribute in vars(self):
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True


@typechecked
@dataclass
class EntityReference:
    name: str
    obj: Union[Entity, None]

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.GET

    @staticmethod
    def type() -> str:
        return "entity"

    def _get(self, stub):
        entityList = stub.GetEntities(iter([pb.Name(name=self.name)]))
        try:
            for entity in entityList:
                self.obj = entity
        except grpc._channel._MultiThreadedRendezvous:
            raise ValueError(f"Entity {self.name} not found.")

    def _get_local(self, db):
        local_entity = db.query_resource("entities", "name", self.name)
        if local_entity == []:
            raise ValueError(f"Entity {self.name} not found.")
        self.obj = local_entity


@typechecked
@dataclass
class ProviderReference:
    name: str
    provider_type: str
    obj: Union[Provider, None]

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.GET

    @staticmethod
    def type() -> str:
        return "provider"

    def _get(self, stub):
        providerList = stub.GetProviders(iter([pb.Name(name=self.name)]))
        try:
            for provider in providerList:
                self.obj = provider
        except grpc._channel._MultiThreadedRendezvous:
            raise ValueError(
                f"Provider {self.name} of type {self.provider_type} not found."
            )

    def _get_local(self, db):
        local_provider = db.query_resource("providers", "name", self.name)
        if local_provider == []:
            raise ValueError("Local mode provider not found.")
        self.obj = local_provider


@typechecked
@dataclass
class SourceReference:
    name: str
    variant: str
    obj: Union[SourceVariant, None]

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.GET

    @staticmethod
    def type() -> str:
        return "source"

    def _get(self, stub):
        sourceList = stub.GetSourceVariants(
            iter([pb.NameVariant(name=self.name, variant=self.variant)])
        )
        try:
            for source in sourceList:
                self.obj = source
        except grpc._channel._MultiThreadedRendezvous:
            raise ValueError(f"Source {self.name}, variant {self.variant} not found.")

    def _get_local(self, db):
        local_source = db.get_source_variant(self.name, self.variant)
        if local_source == []:
            raise ValueError(f"Source {self.name}, variant {self.variant} not found.")
        self.obj = local_source


@typechecked
@dataclass
class TrainingSet:
    name: str
    default_variant: str
    variants: List[str]

    def to_dictionary(self):
        return {
            "name": self.name,
            "default_variant": self.default_variant,
            "variants": self.variants,
        }


@typechecked
@dataclass
class TrainingSetVariant:
    name: str
    owner: str
    label: NameVariant
    features: List[NameVariant]
    description: str
    variant: str
    feature_lags: list = field(default_factory=list)
    tags: list = field(default_factory=list)
    properties: dict = field(default_factory=dict)
    created: str = None
    schedule: str = ""
    schedule_obj: Schedule = None
    provider: str = ""
    status: str = "NO_STATUS"
    error: Optional[str] = None

    def update_schedule(self, schedule) -> None:
        self.schedule_obj = Schedule(
            name=self.name,
            variant=self.variant,
            resource_type=6,
            schedule_string=schedule,
        )
        self.schedule = schedule

    def __post_init__(self):
        if not valid_name_variant(self.label):
            raise ValueError("Label must be set")
        if len(self.features) == 0:
            raise ValueError("A training-set must have atleast one feature")
        for feature in self.features:
            if not valid_name_variant(feature):
                raise ValueError("Invalid Feature")

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def type() -> str:
        return "training-set"

    def get(self, stub):
        name_variant = pb.NameVariant(name=self.name, variant=self.variant)
        ts = next(stub.GetTrainingSetVariants(iter([name_variant])))

        return TrainingSetVariant(
            created=None,
            name=ts.name,
            variant=ts.variant,
            owner=ts.owner,
            description=ts.description,
            status=ts.status.Status._enum_type.values[ts.status.status].name,
            label=(ts.label.name, ts.label.variant),
            features=[(f.name, f.variant) for f in ts.features],
            feature_lags=[],
            provider=ts.provider,
            tags=list(ts.tags.tag),
            properties={k: v for k, v in ts.properties.property.items()},
            error=ts.status.error_message,
        )

    def _create(self, stub) -> None:
        feature_lags = []
        for lag in self.feature_lags:
            lag_duration = Duration()
            _ = lag_duration.FromTimedelta(lag["lag"])
            feature_lag = pb.FeatureLag(
                feature=lag["feature"],
                variant=lag["variant"],
                name=lag["name"],
                lag=lag_duration,
            )
            feature_lags.append(feature_lag)

        serialized = pb.TrainingSetVariant(
            created=None,
            name=self.name,
            variant=self.variant,
            description=self.description,
            schedule=self.schedule,
            owner=self.owner,
            features=[pb.NameVariant(name=v[0], variant=v[1]) for v in self.features],
            label=pb.NameVariant(name=self.label[0], variant=self.label[1]),
            feature_lags=feature_lags,
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
        )
        stub.CreateTrainingSetVariant(serialized)

    def _create_local(self, db) -> None:
        self._check_insert_training_set_resources(db)
        db.insert(
            "training_set_variant",
            str(time.time()),
            self.description,
            self.name,
            self.owner,
            self.variant,
            self.label[0],
            self.label[1],
            "ready",
        )
        if len(self.tags):
            db.upsert(
                "tags",
                self.name,
                self.variant,
                "training_set_variant",
                json.dumps(self.tags),
            )
        if len(self.properties):
            db.upsert(
                "properties",
                self.name,
                self.variant,
                "training_set_variant",
                json.dumps(self.properties),
            )
        self._create_training_set_resource(db)

    def _create_training_set_resource(self, db) -> None:
        db.insert("training_sets", "TrainingSet", self.variant, self.name)

    def _check_insert_training_set_resources(self, db) -> None:
        try:
            db.get_label_variant(self.label[0], self.label[1])
        except ValueError:
            raise LabelNotFound(
                self.label[0], self.label[1], message="Failed to register training set."
            )

        for feature_name, feature_variant in self.features:
            try:
                is_on_demand = db.get_feature_variant_on_demand(
                    feature_name, feature_variant
                )
                if is_on_demand:
                    raise InvalidTrainingSetFeatureComputationMode(
                        feature_name, feature_variant
                    )
            except InvalidTrainingSetFeatureComputationMode as e:
                raise e
            except Exception as e:
                raise FeatureNotFound(
                    feature_name,
                    feature_variant,
                    message=f"Failed to register training set. Error: {e}",
                )

            db.insert(
                "training_set_features",
                self.name,
                self.variant,
                feature_name,  # feature name
                feature_variant,  # feature variant
            )

        for feature in self.feature_lags:
            feature_name = feature["feature"]
            feature_variant = feature["variant"]
            feature_new_name = feature["name"]
            feature_lag = feature["lag"].total_seconds()

            try:
                db.get_feature_variant(feature_name, feature_variant)
            except Exception as e:
                raise FeatureNotFound(
                    feature_name,
                    feature_variant,
                    message=f"Failed to register training set. Error: {e}",
                )

            db.insert(
                "training_set_lag_features",
                self.name,
                self.variant,
                feature_name,  # feature name
                feature_variant,  # feature variant
                feature_new_name,  # feature new name
                feature_lag,  # feature_lag
            )

    def get_status(self):
        return ResourceStatus(self.status)

    def is_ready(self):
        return self.status == ResourceStatus.READY.value

    def __eq__(self, other):
        for attribute in vars(self):
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True


@typechecked
@dataclass
class Model:
    name: str
    description: str = ""
    tags: list = field(default_factory=list)
    properties: dict = field(default_factory=dict)

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    def type(self) -> str:
        return "model"

    def _create(self, stub) -> None:
        properties = pb.Properties(property=self.properties)
        serialized = pb.Model(
            name=self.name,
            tags=pb.Tags(tag=self.tags),
            properties=Properties(self.properties).serialized,
        )
        stub.CreateModel(serialized)

    def _create_local(self, db) -> None:
        db.insert(
            "models",
            self.name,
            "Model",
        )
        if len(self.tags):
            db.upsert("tags", self.name, "", "models", json.dumps(self.tags))
        if len(self.properties):
            db.upsert(
                "properties", self.name, "", "models", json.dumps(self.properties)
            )

    def __eq__(self, other):
        for attribute in vars(self):
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True

    def to_dictionary(self):
        return {
            "name": self.name,
            "description": self.description,
            "tags": self.tags,
            "properties": self.properties,
        }


Resource = Union[
    PrimaryData,
    Provider,
    Entity,
    User,
    FeatureVariant,
    LabelVariant,
    TrainingSetVariant,
    SourceVariant,
    Schedule,
    ProviderReference,
    SourceReference,
    EntityReference,
    Model,
    OnDemandFeatureVariant,
    StreamFeature,
    StreamLabel,
]


class ResourceRedefinedError(Exception):
    @typechecked
    def __init__(self, resource: Resource):
        variantStr = (
            f" variant {resource.variant}" if hasattr(resource, "variant") else ""
        )
        resourceId = f"{resource.name}{variantStr}"
        super().__init__(
            f"{resource.type()} resource {resourceId} defined in multiple places"
        )


class ResourceState:
    def __init__(self):
        self.__state = {}

    def reset(self):
        self.__state = {}

    @typechecked
    def add(self, resource: Resource) -> None:
        if hasattr(resource, "variant"):
            key = (
                resource.operation_type().name,
                resource.type(),
                resource.name,
                resource.variant,
            )
        else:
            key = (resource.operation_type().name, resource.type(), resource.name)
        if key in self.__state:
            if resource == self.__state[key]:
                print(f"Resource {resource.type()} already registered.")
                return
            raise ResourceRedefinedError(resource)
        self.__state[key] = resource
        if hasattr(resource, "schedule_obj") and resource.schedule_obj != None:
            my_schedule = resource.schedule_obj
            key = (my_schedule.type(), my_schedule.name)
            self.__state[key] = my_schedule

    def sorted_list(self) -> List[Resource]:
        resource_order = {
            "user": 0,
            "provider": 1,
            "source": 2,
            "entity": 3,
            "feature": 4,
            "ondemand_feature": 5,
            "label": 6,
            "training-set": 7,
            "schedule": 8,
            "model": 9,
        }

        def to_sort_key(res):
            resource_num = resource_order[res.type()]
            variant = res.variant if hasattr(res, "variant") else ""
            return (resource_num, res.name, variant)

        return sorted(self.__state.values(), key=to_sort_key)

    def create_all_dryrun(self) -> None:
        for resource in self.sorted_list():
            if resource.operation_type() is OperationType.GET:
                print("Getting", resource.type(), resource.name)
            if resource.operation_type() is OperationType.CREATE:
                print("Creating", resource.type(), resource.name)

    def create_all_local(self) -> None:
        db = SQLiteMetadata()
        check_up_to_date(True, "register")
        features = []
        for resource in self.sorted_list():
            if isinstance(resource, FeatureVariant):
                features.append(resource)
            resource_variant = (
                f" {resource.variant}" if hasattr(resource, "variant") else ""
            )
            if resource.operation_type() is OperationType.GET:
                print("Getting", resource.type(), resource.name, resource_variant)
                resource._get_local(db)
            if resource.operation_type() is OperationType.CREATE:
                print("Creating", resource.type(), resource.name, resource_variant)
                resource._create_local(db)
        db.close()
        from .serving import LocalClientImpl

        client = LocalClientImpl()
        for feature in features:
            client.compute_feature(feature.name, feature.variant, feature.entity)
        return

    def create_all(self, stub) -> None:
        check_up_to_date(False, "register")
        for resource in self.sorted_list():
            if resource.type() == "provider" and resource.name == "local-mode":
                continue
            if (
                resource.type() == "feature"
                and getattr(resource, "provider", None) == "local-mode"
            ):
                raise ValueError(
                    f"Inference store must be provided for feature {resource.name} ({resource.variant})"
                )
            try:
                # NOTE: There is an extra space before the variant name to better handle the case
                # where a resource has no variant; ultimately, we should separate data access and
                # logging/CLI output to avoid this unnecessary coupling.
                resource_variant = (
                    f" {resource.variant}" if hasattr(resource, "variant") else ""
                )
                if resource.operation_type() is OperationType.GET:
                    print(
                        f"Getting {resource.type()} {resource.name}{resource_variant}"
                    )
                    resource._get(stub)
                if resource.operation_type() is OperationType.CREATE:
                    print(
                        f"Creating {resource.type()} {resource.name}{resource_variant}"
                    )
                    resource._create(stub)
            except grpc.RpcError as e:
                if e.code() == grpc.StatusCode.ALREADY_EXISTS:
                    print(f"{resource.name}{resource_variant} already exists.")
                    continue

                raise e


## Executor Providers
@typechecked
class DatabricksCredentials:
    def __init__(
        self,
        cluster_id: str,
        username: str = "",
        password: str = "",
        host: str = "",
        token: str = "",
    ):
        """

        Credentials for a Databricks cluster.

        **Example**
        ```
        databricks = ff.DatabricksCredentials(
            username="<my_username>",
            password="<my_password>",
            host="<databricks_hostname>",
            token="<databricks_token>",
            cluster_id="<databricks_cluster>",
        )

        spark = ff.register_spark(
            name="spark",
            executor=databricks,
            ...
        )
        ```

        Args:
            username (str): Username for a Databricks cluster.
            password (str): Password for a Databricks cluster.
            host (str): The hostname of a Databricks cluster.
            token (str): The token for a Databricks cluster.
            cluster_id (str): ID of an existing Databricks cluster.
        """
        self.username = username
        self.password = password
        self.host = host
        self.token = token
        self.cluster_id = cluster_id

        host_token_provided = (
            username == "" and password == "" and host != "" and token != ""
        )
        username_password_provided = (
            username != "" and password != "" and host == "" and token == ""
        )

        if (
            not host_token_provided
            and not username_password_provided
            or host_token_provided
            and username_password_provided
        ):
            raise Exception(
                "The DatabricksCredentials requires only one credentials set ('username' and 'password' or 'host' and 'token' set.)"
            )

        if not cluster_id:
            raise Exception("Cluster_id of existing cluster must be provided")

    def type(self):
        return "DATABRICKS"

    def config(self):
        return {
            "Username": self.username,
            "Password": self.password,
            "Host": self.host,
            "Token": self.token,
            "Cluster": self.cluster_id,
        }


@typechecked
@dataclass
class EMRCredentials:
    def __init__(
        self, emr_cluster_id: str, emr_cluster_region: str, credentials: AWSCredentials
    ):
        """

        Credentials for an EMR cluster.

        **Example**
        ```
        emr = ff.EMRCredentials(
            emr_cluster_id="<cluster_id>",
            emr_cluster_region="<cluster_region>",
            credentials="<AWS_Credentials>",
        )

        spark = ff.register_spark(
            name="spark",
            executor=emr,
            ...
        )
        ```

        Args:
            emr_cluster_id (str): ID of an existing EMR cluster.
            emr_cluster_region (str): Region of an existing EMR cluster.
            credentials (AWSCredentials): Credentials for an AWS account with access to the cluster
        """
        self.emr_cluster_id = emr_cluster_id
        self.emr_cluster_region = emr_cluster_region
        self.credentials = credentials

    def type(self):
        return "EMR"

    def config(self):
        return {
            "ClusterName": self.emr_cluster_id,
            "ClusterRegion": self.emr_cluster_region,
            "Credentials": self.credentials.config(),
        }


@typechecked
@dataclass
class SparkCredentials:
    def __init__(
        self,
        master: str,
        deploy_mode: str,
        python_version: str,
        core_site_path: str = "",
        yarn_site_path: str = "",
    ):
        """

        Credentials for a Generic Spark Cluster

        **Example**
        ```
        spark_credentials = ff.SparkCredentials(
            master="yarn",
            deploy_mode="cluster",
            python_version="3.7.12",
            core_site_path="core-site.xml",
            yarn_site_path="yarn-site.xml"
        )

        spark = ff.register_spark(
            name="spark",
            executor=spark_credentials,
            ...
        )
        ```

        Args:
            master (str): The hostname of the Spark cluster. (The same that would be passed to `spark-submit`).
            deploy_mode (str): The deploy mode of the Spark cluster. (The same that would be passed to `spark-submit`).
            python_version (str): The Python version running on the cluster. Supports 3.7-3.11
            core_site_path (str): The path to the core-site.xml file. (For Yarn clusters only)
            yarn_site_path (str): The path to the yarn-site.xml file. (For Yarn clusters only)
        """
        self.master = master.lower()
        self.deploy_mode = deploy_mode.lower()
        self.core_site_path = core_site_path
        self.yarn_site_path = yarn_site_path

        if self.deploy_mode != "cluster" and self.deploy_mode != "client":
            raise Exception(
                f"Spark does not support '{self.deploy_mode}' deploy mode. It only supports 'cluster' and 'client'."
            )

        self.python_version = self._verify_python_version(
            self.deploy_mode, python_version
        )

        self._verify_yarn_config()

    def _verify_python_version(self, deploy_mode, version):
        if deploy_mode == "cluster" and version == "":
            client_python_version = sys.version_info
            client_major = str(client_python_version.major)
            client_minor = str(client_python_version.minor)

            if client_major != MAJOR_VERSION:
                client_major = "3"
            if client_minor not in MINOR_VERSIONS:
                client_minor = "7"

            version = f"{client_major}.{client_minor}"

        if version.count(".") == 2:
            major, minor, _ = version.split(".")
        elif version.count(".") == 1:
            major, minor = version.split(".")
        else:
            raise Exception(
                "Please specify your Python version on the Spark cluster. Accepted formats: Major.Minor or Major.Minor.Patch; ex. '3.7' or '3.7.16"
            )

        if major != MAJOR_VERSION or minor not in MINOR_VERSIONS:
            raise Exception(
                f"The Python version {version} is not supported. Currently, supported versions are 3.7-3.10."
            )

        """
        The Python versions on the Docker image are 3.7.16, 3.8.16, 3.9.16, 3.10.10, and 3.11.2.
        This conditional statement sets the patch number based on the minor version. 
        """
        if minor == "10":
            patch = "10"
        elif minor == "11":
            patch = "2"
        else:
            patch = "16"

        return f"{major}.{minor}.{patch}"

    def _verify_yarn_config(self):
        if self.master == "yarn" and (
            self.core_site_path == "" or self.yarn_site_path == ""
        ):
            raise Exception(
                "Yarn requires core-site.xml and yarn-site.xml files."
                "Please copy these files from your Spark instance to local, then provide the local path in "
                "core_site_path and yarn_site_path. "
            )

    def type(self):
        return "SPARK"

    def config(self):
        core_site = (
            "" if self.core_site_path == "" else open(self.core_site_path, "r").read()
        )
        yarn_site = (
            "" if self.yarn_site_path == "" else open(self.yarn_site_path, "r").read()
        )

        return {
            "Master": self.master,
            "DeployMode": self.deploy_mode,
            "PythonVersion": self.python_version,
            "CoreSite": core_site,
            "YarnSite": yarn_site,
        }


@typechecked
@dataclass
class TrainingSetFeatures:
    training_set_name: str
    training_set_variant: str
    feature_name: str
    feature_variant: str

    def to_dictionary(self):
        return {
            "training_set_name": self.training_set_name,
            "training_set_variant": self.training_set_variant,
            "feature_name": self.feature_name,
            "feature_variant": self.feature_variant,
        }


ExecutorCredentials = Union[EMRCredentials, DatabricksCredentials, SparkCredentials]
